from typing import Dict, List

from moto.core.base_backend import BackendDict, BaseBackend
from moto.utilities.paginator import paginate

from .data_models import (
    QuicksightDataSet,
    QuicksightGroup,
    QuicksightIngestion,
    QuicksightMembership,
    QuicksightUser,
)
from .exceptions import ResourceNotFoundException
from .utils import PAGINATION_MODEL, QuicksightSearchFilterFactory


def _create_id(aws_account_id: str, namespace: str, _id: str) -> str:
    return f"{aws_account_id}:{namespace}:{_id}"


class QuickSightBackend(BaseBackend):
    """Implementation of QuickSight APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.groups: Dict[str, QuicksightGroup] = dict()
        self.users: Dict[str, QuicksightUser] = dict()

    def create_data_set(self, data_set_id: str, name: str) -> QuicksightDataSet:
        return QuicksightDataSet(
            self.account_id, self.region_name, data_set_id, name=name
        )

    def create_group(
        self, group_name: str, description: str, aws_account_id: str, namespace: str
    ) -> QuicksightGroup:
        group = QuicksightGroup(
            region=self.region_name,
            group_name=group_name,
            description=description,
            aws_account_id=aws_account_id,
            namespace=namespace,
        )
        _id = _create_id(aws_account_id, namespace, group_name)
        self.groups[_id] = group
        return group

    def create_group_membership(
        self, aws_account_id: str, namespace: str, group_name: str, member_name: str
    ) -> QuicksightMembership:
        group = self.describe_group(aws_account_id, namespace, group_name)
        return group.add_member(member_name)

    def create_ingestion(
        self, data_set_id: str, ingestion_id: str
    ) -> QuicksightIngestion:
        return QuicksightIngestion(
            self.account_id, self.region_name, data_set_id, ingestion_id
        )

    def delete_group(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> None:
        _id = _create_id(aws_account_id, namespace, group_name)
        self.groups.pop(_id, None)

    def delete_user(self, aws_account_id: str, namespace: str, user_name: str) -> None:
        # Delete users from all groups
        for group in self.groups.values():
            group.delete_member(user_name)
        # Delete user itself
        _id = _create_id(aws_account_id, namespace, user_name)
        self.users.pop(_id, None)

    def describe_group(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> QuicksightGroup:
        _id = _create_id(aws_account_id, namespace, group_name)
        if _id not in self.groups:
            raise ResourceNotFoundException(f"Group {group_name} not found")
        return self.groups[_id]

    def describe_group_membership(
        self, aws_account_id: str, namespace: str, group_name: str, member_name: str
    ) -> QuicksightMembership:
        group = self.describe_group(aws_account_id, namespace, group_name)
        member = group.get_member(member_name)
        if member is None:
            raise ResourceNotFoundException(f"Member {member_name} not found")
        return member

    def describe_user(
        self, aws_account_id: str, namespace: str, user_name: str
    ) -> QuicksightUser:
        _id = _create_id(aws_account_id, namespace, user_name)
        if _id not in self.users:
            raise ResourceNotFoundException(f"User {user_name} not found")
        return self.users[_id]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_groups(self, aws_account_id: str, namespace: str) -> List[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        return [
            group for _id, group in self.groups.items() if _id.startswith(id_for_ns)
        ]

    @paginate(pagination_model=PAGINATION_MODEL)
    def search_groups(
        self, aws_account_id: str, namespace: str, filters: List[Dict[str, str]]
    ) -> List[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        filter_list = QuicksightSearchFilterFactory.validate_and_create_filter(
            model_type=QuicksightGroup, input=filters
        )
        return [
            group
            for _id, group in self.groups.items()
            if _id.startswith(id_for_ns) and filter_list.match(group)
        ]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_group_memberships(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> List[QuicksightMembership]:
        group = self.describe_group(aws_account_id, namespace, group_name)
        return group.list_members()

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_users(self, aws_account_id: str, namespace: str) -> List[QuicksightUser]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        return [user for _id, user in self.users.items() if _id.startswith(id_for_ns)]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_user_groups(
        self, aws_account_id: str, namespace: str, user_name: str
    ) -> List[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        group_list: Dict[str, QuicksightGroup] = {}
        # Loop through all groups and check if the user is member.
        for id, group in self.groups.items():
            if group.get_member(user_name):
                group_list[id] = group
        return [group for _id, group in group_list.items() if _id.startswith(id_for_ns)]

    def register_user(
        self,
        identity_type: str,
        email: str,
        user_role: str,
        aws_account_id: str,
        namespace: str,
        user_name: str,
    ) -> QuicksightUser:
        """
        The following parameters are not yet implemented:
        IamArn, SessionName, CustomsPermissionsName, ExternalLoginFederationProviderType, CustomFederationProviderUrl, ExternalLoginId
        """
        user = QuicksightUser(
            account_id=aws_account_id,
            region=self.region_name,
            email=email,
            identity_type=identity_type,
            user_role=user_role,
            username=user_name,
        )
        _id = _create_id(aws_account_id, namespace, user_name)
        self.users[_id] = user
        return user

    def update_user(
        self,
        aws_account_id: str,
        namespace: str,
        user_name: str,
        email: str,
        user_role: str,
    ) -> QuicksightUser:
        user = self.describe_user(aws_account_id, namespace, user_name)
        user.email = email
        user.user_role = user_role
        return user

    def update_group(
        self, aws_account_id: str, namespace: str, group_name: str, description: str
    ) -> QuicksightGroup:
        group = self.describe_group(aws_account_id, namespace, group_name)
        group.description = description
        return group


quicksight_backends = BackendDict(QuickSightBackend, "quicksight")



"""identitystore base URL and path."""

from .responses import IdentityStoreResponse

url_bases = [
    r"https?://identitystore\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": IdentityStoreResponse.dispatch,
}



import boto3
import pytest
from boto3 import Session
from botocore.client import ClientError

from moto import mock_aws


@mock_aws
def test_get_public_access_block_for_account():
    from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

    client = boto3.client("s3control", region_name="us-west-2")

    # With an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId="111111111111")
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Without one defined:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId=ACCOUNT_ID)
    assert (
        ce_err.value.response["Error"]["Code"] == "NoSuchPublicAccessBlockConfiguration"
    )

    # Put a with an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.put_public_access_block(
            AccountId="111111111111",
            PublicAccessBlockConfiguration={"BlockPublicAcls": True},
        )
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Put with an invalid PAB:
    with pytest.raises(ClientError) as ce_err:
        client.put_public_access_block(
            AccountId=ACCOUNT_ID, PublicAccessBlockConfiguration={}
        )
    assert ce_err.value.response["Error"]["Code"] == "InvalidRequest"
    assert (
        "Must specify at least one configuration."
        in ce_err.value.response["Error"]["Message"]
    )

    # Correct PAB:
    client.put_public_access_block(
        AccountId=ACCOUNT_ID,
        PublicAccessBlockConfiguration={
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        },
    )

    # Get the correct PAB (for all regions):
    for region in Session().get_available_regions("s3control"):
        region_client = boto3.client("s3control", region_name=region)
        assert region_client.get_public_access_block(AccountId=ACCOUNT_ID)[
            "PublicAccessBlockConfiguration"
        ] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        }

    # Delete with an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.delete_public_access_block(AccountId="111111111111")
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Delete successfully:
    client.delete_public_access_block(AccountId=ACCOUNT_ID)

    # Confirm that it's deleted:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId=ACCOUNT_ID)
    assert (
        ce_err.value.response["Error"]["Code"] == "NoSuchPublicAccessBlockConfiguration"
    )



import abc
import copy
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.intrinsic.function.function import Function
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.parse.intrinsic.intrinsic_parser import (
    IntrinsicParser,
)
from moto.stepfunctions.parser.asl.utils.json_path import extract_json

_CREDENTIALS_ROLE_ARN_KEY: str = "RoleArn"
ComputedCredentials = dict


class RoleArn(EvalComponent, abc.ABC): ...


class RoleArnConst(RoleArn):
    value: str

    def __init__(self, value: str):
        self.value = value

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self.value)


class RoleArnJSONata(RoleArn):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def _eval_body(self, env: Environment) -> None:
        self.jsonata_template_value_terminal_expression.eval(env=env)


class RoleArnVar(RoleArn):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        self.variable_sample = variable_sample

    def _eval_body(self, env: Environment) -> None:
        self.variable_sample.eval(env=env)


class RoleArnPath(RoleArnConst):
    def _eval_body(self, env: Environment) -> None:
        current_output = env.stack[-1]
        arn = extract_json(self.value, current_output)
        env.stack.append(arn)


class RoleArnContextObject(RoleArnConst):
    def _eval_body(self, env: Environment) -> None:
        value = extract_json(self.value, env.states.context_object.context_object_data)
        env.stack.append(copy.deepcopy(value))


class RoleArnIntrinsicFunction(RoleArnConst):
    function: Function

    def __init__(self, value: str) -> None:
        super().__init__(value=value)
        self.function, _ = IntrinsicParser.parse(value)

    def _eval_body(self, env: Environment) -> None:
        self.function.eval(env=env)


class Credentials(EvalComponent):
    role_arn: RoleArn

    def __init__(self, role_arn: RoleArn):
        self.role_arn = role_arn

    @staticmethod
    def get_role_arn_from(computed_credentials: ComputedCredentials) -> Optional[str]:
        return computed_credentials.get(_CREDENTIALS_ROLE_ARN_KEY)

    def _eval_body(self, env: Environment) -> None:
        self.role_arn.eval(env=env)
        role_arn = env.stack.pop()
        computes_credentials: ComputedCredentials = {
            _CREDENTIALS_ROLE_ARN_KEY: role_arn
        }
        env.stack.append(computes_credentials)



import json
import string
from typing import Any, Dict, List, Union

import yaml

from moto.moto_api._internal import mock_random as random
from moto.utilities.id_generator import ResourceIdentifier, Tags, generate_str_id


class ApigwIdentifier(ResourceIdentifier):
    service = "apigateway"

    def __init__(self, account_id: str, region: str, name: str):
        super().__init__(account_id, region, name)

    def generate(
        self, existing_ids: Union[List[str], None] = None, tags: Tags = None
    ) -> str:
        return generate_str_id(
            resource_identifier=self,
            existing_ids=existing_ids,
            tags=tags,
            length=10,
            include_digits=True,
            lower_case=True,
        )


class ApigwApiKeyIdentifier(ApigwIdentifier):
    resource = "api_key"

    def __init__(self, account_id: str, region: str, value: str):
        super().__init__(account_id, region, value)


class ApigwAuthorizerIdentifier(ApigwIdentifier):
    resource = "authorizer"


class ApigwDeploymentIdentifier(ApigwIdentifier):
    resource = "deployment"

    def __init__(self, account_id: str, region: str, stage_name: str):
        super().__init__(account_id, region, stage_name)


class ApigwModelIdentifier(ApigwIdentifier):
    resource = "model"


class ApigwRequestValidatorIdentifier(ApigwIdentifier):
    resource = "request_validator"


class ApigwResourceIdentifier(ApigwIdentifier):
    resource = "resource"

    def __init__(
        self, account_id: str, region: str, parent_id: str = "", path_name: str = "/"
    ):
        super().__init__(
            account_id,
            region,
            ".".join((parent_id, path_name)),
        )


class ApigwRestApiIdentifier(ApigwIdentifier):
    resource = "rest_api"


class ApigwUsagePlanIdentifier(ApigwIdentifier):
    resource = "usage_plan"


class ApigwVpcLinkIdentifier(ApigwIdentifier):
    resource = "vpc_link"


def create_id() -> str:
    size = 10
    chars = list(range(10)) + list(string.ascii_lowercase)
    return "".join(str(random.choice(chars)) for x in range(size))


def deserialize_body(body: str) -> Dict[str, Any]:
    try:
        api_doc = json.loads(body)
    except json.JSONDecodeError:
        api_doc = yaml.safe_load(body)

    if "openapi" in api_doc or "swagger" in api_doc:
        return api_doc

    return {}


def to_path(prop: str) -> str:
    return "/" + prop



"""Unit tests for appmesh-supported APIs."""

from collections import defaultdict
from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

from .data import (
    grpc_route_spec,
    grpc_virtual_node_spec,
    http2_route_spec,
    http2_virtual_node_spec,
    http_route_spec,
    http_virtual_node_spec,
    modified_http2_virtual_node_spec,
    modified_http_route_spec,
    tcp_route_spec,
    tcp_virtual_node_spec,
)

MESH_NAME = "mock_mesh"


@pytest.fixture(name="client")
def fixture_transfer_client():
    with mock_aws():
        yield boto3.client("appmesh", region_name="us-east-1")


@mock_aws
def test_create_list_update_describe_delete_mesh(client):
    connection = client.create_mesh(
        meshName="mesh1",
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    mesh = connection.get("mesh")
    assert mesh["meshName"] == "mesh1"
    assert mesh["spec"]["egressFilter"]["type"] == "DROP_ALL"
    assert mesh["spec"]["serviceDiscovery"]["ipPreference"] == "IPv4_ONLY"
    assert mesh["status"]["status"] == "ACTIVE"
    assert mesh["metadata"]["version"] == 1
    assert isinstance(mesh["metadata"]["arn"], str)
    assert isinstance(mesh["metadata"]["meshOwner"], str)
    assert isinstance(mesh["metadata"]["resourceOwner"], str)
    assert isinstance(mesh["metadata"]["createdAt"], datetime)
    assert isinstance(mesh["metadata"]["lastUpdatedAt"], datetime)

    connection = client.create_mesh(
        meshName="mesh2",
        spec={
            "egressFilter": {"type": "ALLOW_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_PREFERRED"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    mesh = connection.get("mesh")
    assert mesh["meshName"] == "mesh2"
    assert mesh["spec"]["egressFilter"]["type"] == "ALLOW_ALL"
    assert mesh["spec"]["serviceDiscovery"]["ipPreference"] == "IPv4_PREFERRED"
    assert mesh["status"]["status"] == "ACTIVE"
    assert mesh["metadata"]["version"] == 1
    assert isinstance(mesh["metadata"]["arn"], str)
    assert isinstance(mesh["metadata"]["meshOwner"], str)
    assert isinstance(mesh["metadata"]["resourceOwner"], str)
    assert isinstance(mesh["metadata"]["createdAt"], datetime)
    assert isinstance(mesh["metadata"]["lastUpdatedAt"], datetime)

    connection = client.list_meshes()
    meshes = connection.get("meshes")
    assert isinstance(meshes, list)
    assert len(meshes) == 2
    names_counted = defaultdict(int)
    for mesh in meshes:
        mesh_name = mesh.get("meshName")
        if mesh_name:
            names_counted[mesh_name] += 1
    assert names_counted["mesh1"] == 1
    assert names_counted["mesh2"] == 1

    connection = client.update_mesh(
        meshName="mesh1",
        spec={
            "egressFilter": {"type": "ALLOW_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv6_PREFERRED"},
        },
    )
    mesh = connection.get("mesh")
    assert mesh["meshName"] == "mesh1"
    assert mesh["spec"]["egressFilter"]["type"] == "ALLOW_ALL"
    assert mesh["spec"]["serviceDiscovery"]["ipPreference"] == "IPv6_PREFERRED"
    assert mesh["status"]["status"] == "ACTIVE"
    assert mesh["metadata"]["version"] == 2

    connection = client.describe_mesh(meshName="mesh1")
    mesh = connection.get("mesh")
    assert mesh["meshName"] == "mesh1"
    assert mesh["spec"]["egressFilter"]["type"] == "ALLOW_ALL"
    assert mesh["spec"]["serviceDiscovery"]["ipPreference"] == "IPv6_PREFERRED"
    assert mesh["status"]["status"] == "ACTIVE"
    assert mesh["metadata"]["version"] == 2

    connection = client.delete_mesh(meshName="mesh2")
    mesh = connection.get("mesh")
    assert mesh["meshName"] == "mesh2"
    assert mesh["spec"]["egressFilter"]["type"] == "ALLOW_ALL"
    assert mesh["spec"]["serviceDiscovery"]["ipPreference"] == "IPv4_PREFERRED"
    assert mesh["status"]["status"] == "DELETED"

    connection = client.list_meshes()
    meshes = connection.get("meshes")
    assert isinstance(meshes, list)
    assert len(meshes) == 1
    assert meshes[0]["meshName"] == "mesh1"

    with pytest.raises(ClientError) as e:
        client.describe_mesh(meshName="mesh2")
    err = e.value.response["Error"]
    assert err["Code"] == "MeshNotFound"
    assert err["Message"] == "There are no meshes with the name mesh2."


@mock_aws
def test_list_meshes_paginated(client):
    for i in range(5):
        client.create_mesh(
            meshName=f"mesh{i}",
            spec={
                "egressFilter": {"type": "DROP_ALL"},
                "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
            },
            tags=[{"key": "owner", "value": "moto"}],
        )

    all_meshes = client.list_meshes()["meshes"]
    assert len(all_meshes) == 5

    page1 = client.list_meshes(limit=2)
    assert len(page1["meshes"]) == 2

    page2 = client.list_meshes(nextToken=page1["nextToken"])
    assert len(page2["meshes"]) == 3


@mock_aws
def test_create_describe_list_update_delete_virtual_router(client):
    connection = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    assert "mesh" in connection
    mesh = connection["mesh"]
    assert "metadata" in mesh
    mesh_owner = mesh["metadata"]["meshOwner"]

    ROUTER_1 = "router1"
    ROUTER_2 = "router2"

    connection = client.create_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_1,
        spec={"listeners": [{"portMapping": {"port": 80, "protocol": "http"}}]},
        tags=[{"key": "router_traffic", "value": "http"}],
    )
    router1 = connection.get("virtualRouter")
    assert router1["meshName"] == MESH_NAME
    assert router1["metadata"]["meshOwner"] == mesh_owner
    assert router1["metadata"]["version"] == 1
    assert router1["spec"]["listeners"][0]["portMapping"]["port"] == 80
    assert router1["spec"]["listeners"][0]["portMapping"]["protocol"] == "http"
    assert router1["status"]["status"] == "ACTIVE"

    connection = client.create_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_2,
        spec={"listeners": [{"portMapping": {"port": 443, "protocol": "http2"}}]},
        tags=[{"key": "router_traffic", "value": "https"}],
    )
    router2 = connection.get("virtualRouter")
    assert router2["meshName"] == MESH_NAME
    assert router2["metadata"]["meshOwner"] == mesh_owner
    assert router2["metadata"]["version"] == 1
    assert router2["spec"]["listeners"][0]["portMapping"]["port"] == 443
    assert router2["spec"]["listeners"][0]["portMapping"]["protocol"] == "http2"
    assert router2["status"]["status"] == "ACTIVE"
    connection = client.list_virtual_routers(meshName=MESH_NAME, meshOwner=mesh_owner)
    virtual_routers = connection.get("virtualRouters")
    assert isinstance(virtual_routers, list)
    assert len(virtual_routers) == 2
    names_counted = defaultdict(int)
    for virtual_router in virtual_routers:
        router_name = virtual_router.get("virtualRouterName")
        if router_name:
            names_counted[router_name] += 1
    assert names_counted[ROUTER_1] == 1
    assert names_counted[ROUTER_2] == 1

    connection = client.update_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_2,
        spec={"listeners": [{"portMapping": {"port": 80, "protocol": "tcp"}}]},
    )
    updated_router2 = connection.get("virtualRouter")
    assert updated_router2["virtualRouterName"] == ROUTER_2
    assert updated_router2["meshName"] == MESH_NAME
    assert updated_router2["metadata"]["meshOwner"] == mesh_owner
    assert updated_router2["metadata"]["version"] == 2
    assert updated_router2["spec"]["listeners"][0]["portMapping"]["port"] == 80
    assert updated_router2["spec"]["listeners"][0]["portMapping"]["protocol"] == "tcp"
    assert updated_router2["status"]["status"] == "ACTIVE"

    connection = client.describe_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_2,
    )
    described_router2 = connection.get("virtualRouter")
    assert described_router2["virtualRouterName"] == ROUTER_2
    assert described_router2["meshName"] == MESH_NAME
    assert described_router2["metadata"]["meshOwner"] == mesh_owner
    assert described_router2["metadata"]["version"] == 2
    assert described_router2["spec"]["listeners"][0]["portMapping"]["port"] == 80
    assert described_router2["spec"]["listeners"][0]["portMapping"]["protocol"] == "tcp"
    assert described_router2["status"]["status"] == "ACTIVE"

    connection = client.delete_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_1,
    )
    deleted_router1 = connection.get("virtualRouter")
    assert deleted_router1["virtualRouterName"] == ROUTER_1
    assert deleted_router1["meshName"] == MESH_NAME
    assert deleted_router1["metadata"]["meshOwner"] == mesh_owner
    assert deleted_router1["metadata"]["version"] == 1
    assert deleted_router1["spec"]["listeners"][0]["portMapping"]["port"] == 80
    assert deleted_router1["spec"]["listeners"][0]["portMapping"]["protocol"] == "http"
    assert deleted_router1["status"]["status"] == "DELETED"
    with pytest.raises(ClientError) as e:
        client.describe_virtual_router(
            meshName=MESH_NAME,
            meshOwner=mesh_owner,
            virtualRouterName=ROUTER_1,
        )
    err = e.value.response["Error"]
    assert err["Code"] == "VirtualRouterNotFound"
    assert (
        err["Message"]
        == f"The mesh {MESH_NAME} does not have a virtual router named {ROUTER_1}."
    )


@mock_aws
def test_list_virtual_routers_paginated(client):
    mesh = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )["mesh"]
    mesh_owner = mesh["metadata"]["meshOwner"]

    for i in range(5):
        client.create_virtual_router(
            meshName=MESH_NAME,
            meshOwner=mesh_owner,
            virtualRouterName=f"router{i}",
            spec={"listeners": [{"portMapping": {"port": 80, "protocol": "http"}}]},
            tags=[{"key": "router_traffic", "value": "http"}],
        )

    all_routers = client.list_virtual_routers(meshName=MESH_NAME, meshOwner=mesh_owner)[
        "virtualRouters"
    ]
    assert len(all_routers) == 5

    page1 = client.list_virtual_routers(
        meshName=MESH_NAME, meshOwner=mesh_owner, limit=2
    )
    assert len(page1["virtualRouters"]) == 2

    page2 = client.list_virtual_routers(
        meshName=MESH_NAME, meshOwner=mesh_owner, nextToken=page1["nextToken"]
    )
    assert len(page2["virtualRouters"]) == 3


@mock_aws
def test_create_describe_list_update_delete_route(client):
    ROUTER_NAME = "mock_virtual_router"

    connection = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    assert "mesh" in connection
    mesh = connection["mesh"]
    assert "metadata" in mesh
    mesh_owner = mesh["metadata"]["meshOwner"]

    connection = client.create_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_NAME,
        spec={"listeners": [{"portMapping": {"port": 80, "protocol": "http"}}]},
        tags=[{"key": "router_traffic", "value": "http"}],
    )

    router = connection.get("virtualRouter")
    assert router["meshName"] == MESH_NAME
    assert router["metadata"]["meshOwner"] == mesh_owner
    assert router["virtualRouterName"] == ROUTER_NAME

    ROUTE_1 = "route1"
    ROUTE_2 = "route2"
    ROUTE_3 = "route3"
    ROUTE_4 = "route4"

    connection = client.create_route(
        meshOwner=mesh_owner,
        meshName=MESH_NAME,
        virtualRouterName=ROUTER_NAME,
        routeName=ROUTE_1,
        tags=[{"key": "license", "value": "apache"}],
        spec=grpc_route_spec,
    )

    route = connection["route"]

    spec = route["spec"]
    assert spec["priority"] == 1
    grpc_route = spec["grpcRoute"]

    assert len(grpc_route["action"]["weightedTargets"]) == 1
    weighted_target = grpc_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 8080
    assert weighted_target["virtualNode"] == "my-virtual-node"
    assert weighted_target["weight"] == 50

    assert len(grpc_route["match"]["metadata"]) == 1
    metadata = grpc_route["match"]["metadata"][0]
    assert metadata["invert"] is False
    assert metadata["match"]["exact"] == "example-value"
    assert metadata["name"] == "my-metadata-key"
    assert grpc_route["match"]["methodName"] == "myMethod"
    assert grpc_route["match"]["port"] == 8080
    assert grpc_route["match"]["serviceName"] == "myService"

    assert grpc_route["retryPolicy"]["grpcRetryEvents"] == [
        "unavailable",
        "resource-exhausted",
    ]
    assert grpc_route["retryPolicy"]["httpRetryEvents"] == ["gateway-error"]
    assert grpc_route["retryPolicy"]["maxRetries"] == 3
    assert grpc_route["retryPolicy"]["perRetryTimeout"]["unit"] == "ms"
    assert grpc_route["retryPolicy"]["perRetryTimeout"]["value"] == 200
    assert grpc_route["retryPolicy"]["tcpRetryEvents"] == ["connection-error"]

    assert grpc_route["timeout"]["idle"]["unit"] == "s"
    assert grpc_route["timeout"]["idle"]["value"] == 60
    assert grpc_route["timeout"]["perRequest"]["unit"] == "s"
    assert grpc_route["timeout"]["perRequest"]["value"] == 5

    connection = client.create_route(
        meshOwner=mesh_owner,
        meshName=MESH_NAME,
        virtualRouterName=ROUTER_NAME,
        routeName=ROUTE_2,
        tags=[{"key": "license", "value": "mit"}],
        spec=http_route_spec,
    )

    route = connection["route"]

    spec = route["spec"]
    assert spec["priority"] == 2
    http_route = http_route_spec["httpRoute"]

    assert len(http_route["action"]["weightedTargets"]) == 1
    weighted_target = http_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 80
    assert weighted_target["virtualNode"] == "web-server-node"
    assert weighted_target["weight"] == 100

    assert len(http_route["match"]["headers"]) == 1
    header = http_route["match"]["headers"][0]
    assert header["invert"] is True
    assert header["match"]["prefix"] == "Bearer "
    assert header["name"] == "Authorization"
    assert http_route["match"]["method"] == "POST"
    assert http_route["match"]["path"]["exact"] == "/login"
    assert http_route["match"]["port"] == 80
    assert (
        http_route["match"]["queryParameters"][0]["match"]["exact"] == "example-match"
    )
    assert http_route["match"]["queryParameters"][0]["name"] == "http-query-param"
    assert http_route["match"]["scheme"] == "http"

    assert http_route["retryPolicy"]["httpRetryEvents"] == [
        "gateway-error",
        "client-error",
    ]
    assert http_route["retryPolicy"]["maxRetries"] == 0
    assert http_route["retryPolicy"]["perRetryTimeout"]["unit"] == "ms"
    assert http_route["retryPolicy"]["perRetryTimeout"]["value"] == 0
    assert http_route["retryPolicy"]["tcpRetryEvents"] == ["connection-error"]

    assert http_route["timeout"]["idle"]["unit"] == "s"
    assert http_route["timeout"]["idle"]["value"] == 15
    assert http_route["timeout"]["perRequest"]["unit"] == "s"
    assert http_route["timeout"]["perRequest"]["value"] == 1

    connection = client.create_route(
        meshOwner=mesh_owner,
        meshName=MESH_NAME,
        virtualRouterName=ROUTER_NAME,
        routeName=ROUTE_3,
        tags=[{"key": "license", "value": "mpl"}],
        spec=http2_route_spec,
    )

    route = connection["route"]

    spec = route["spec"]
    assert spec["priority"] == 3
    http2_route = spec["http2Route"]

    assert len(http2_route["action"]["weightedTargets"]) == 1
    weighted_target = http2_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 80
    assert weighted_target["virtualNode"] == "web-server-node"
    assert weighted_target["weight"] == 75

    assert len(http2_route["match"]["headers"]) == 1
    header = http2_route["match"]["headers"][0]
    assert header["invert"] is False
    assert header["match"]["exact"] == "application/json"
    assert header["name"] == "Content-Type"
    assert http2_route["match"]["method"] == "GET"
    assert http2_route["match"]["path"]["exact"] == "/api/products"
    assert http2_route["match"]["port"] == 80
    assert http2_route["match"]["prefix"] == "/api"
    assert len(http2_route["match"]["queryParameters"]) == 1
    query_param = http2_route["match"]["queryParameters"][0]
    assert query_param["match"]["exact"] == "electronics"
    assert query_param["name"] == "category"
    assert http2_route["match"]["scheme"] == "https"

    assert http2_route["retryPolicy"]["httpRetryEvents"] == ["server-error"]
    assert http2_route["retryPolicy"]["maxRetries"] == 2
    assert http2_route["retryPolicy"]["perRetryTimeout"]["unit"] == "ms"
    assert http2_route["retryPolicy"]["perRetryTimeout"]["value"] == 500
    assert http2_route["retryPolicy"]["tcpRetryEvents"] == ["connection-error"]

    assert http2_route["timeout"]["idle"]["unit"] == "s"
    assert http2_route["timeout"]["idle"]["value"] == 30
    assert http2_route["timeout"]["perRequest"]["unit"] == "s"
    assert http2_route["timeout"]["perRequest"]["value"] == 2

    connection = client.create_route(
        meshOwner=mesh_owner,
        meshName=MESH_NAME,
        virtualRouterName=ROUTER_NAME,
        routeName=ROUTE_4,
        tags=[{"key": "license", "value": "bsd"}],
        spec=tcp_route_spec,
    )

    route = connection["route"]

    spec = route["spec"]
    assert spec["priority"] == 4
    tcp_route = tcp_route_spec["tcpRoute"]

    assert len(tcp_route["action"]["weightedTargets"]) == 1
    weighted_target = tcp_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 22
    assert weighted_target["virtualNode"] == "ssh-server-node"
    assert weighted_target["weight"] == 100

    assert tcp_route["match"]["port"] == 22

    assert tcp_route["timeout"]["idle"]["unit"] == "s"
    assert tcp_route["timeout"]["idle"]["value"] == 600

    connection = client.list_routes(meshName=MESH_NAME, virtualRouterName=ROUTER_NAME)
    routes = connection.get("routes")
    assert isinstance(routes, list)
    assert len(routes) == 4
    names_counted = defaultdict(int)
    for route in routes:
        route_name = route.get("routeName")
        if route_name:
            names_counted[route_name] += 1
    assert names_counted[ROUTE_1] == 1
    assert names_counted[ROUTE_2] == 1
    assert names_counted[ROUTE_3] == 1
    assert names_counted[ROUTE_4] == 1

    connection = client.update_route(
        meshName=MESH_NAME,
        routeName=ROUTE_2,
        virtualRouterName=ROUTER_NAME,
        spec=modified_http_route_spec,
    )

    route = connection["route"]
    assert route["metadata"]["version"] == 2

    spec = route["spec"]
    assert spec["priority"] == 5

    modified_http_route = spec["httpRoute"]

    assert len(modified_http_route["action"]["weightedTargets"]) == 1
    weighted_target = modified_http_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 8080
    assert weighted_target["virtualNode"] == "api-server-node"
    assert weighted_target["weight"] == 50

    assert len(modified_http_route["match"]["headers"]) == 1
    header = modified_http_route["match"]["headers"][0]
    assert header["invert"] is False
    assert header["match"]["prefix"] == "Token "
    assert header["name"] == "X-Auth-Token"
    assert modified_http_route["match"]["method"] == "GET"
    assert modified_http_route["match"]["path"]["exact"] == "/profile"
    assert modified_http_route["match"]["port"] == 443
    assert len(modified_http_route["match"]["queryParameters"]) == 1
    query_param = modified_http_route["match"]["queryParameters"][0]
    assert query_param["match"]["exact"] == "modified-match"
    assert query_param["name"] == "filter-param"
    assert modified_http_route["match"]["scheme"] == "https"

    assert modified_http_route["retryPolicy"]["httpRetryEvents"] == ["server-error"]
    assert modified_http_route["retryPolicy"]["maxRetries"] == 3
    assert modified_http_route["retryPolicy"]["perRetryTimeout"]["unit"] == "s"
    assert modified_http_route["retryPolicy"]["perRetryTimeout"]["value"] == 2
    assert modified_http_route["retryPolicy"]["tcpRetryEvents"] == ["connection-reset"]

    assert modified_http_route["timeout"]["idle"]["unit"] == "m"
    assert modified_http_route["timeout"]["idle"]["value"] == 5
    assert modified_http_route["timeout"]["perRequest"]["unit"] == "ms"
    assert modified_http_route["timeout"]["perRequest"]["value"] == 500
    connection = client.describe_route(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        routeName=ROUTE_2,
        virtualRouterName=ROUTER_NAME,
    )

    route = connection["route"]
    assert route["metadata"]["version"] == 2

    spec = route["spec"]
    assert spec["priority"] == 5

    described_http_route = spec["httpRoute"]

    assert len(described_http_route["action"]["weightedTargets"]) == 1
    weighted_target = described_http_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 8080
    assert weighted_target["virtualNode"] == "api-server-node"
    assert weighted_target["weight"] == 50

    assert len(described_http_route["match"]["headers"]) == 1
    header = described_http_route["match"]["headers"][0]
    assert header["invert"] is False
    assert header["match"]["prefix"] == "Token "
    assert header["name"] == "X-Auth-Token"
    assert described_http_route["match"]["method"] == "GET"
    assert described_http_route["match"]["path"]["exact"] == "/profile"
    assert described_http_route["match"]["port"] == 443
    assert len(described_http_route["match"]["queryParameters"]) == 1
    query_param = described_http_route["match"]["queryParameters"][0]
    assert query_param["match"]["exact"] == "modified-match"
    assert query_param["name"] == "filter-param"
    assert described_http_route["match"]["scheme"] == "https"

    assert described_http_route["retryPolicy"]["httpRetryEvents"] == ["server-error"]
    assert described_http_route["retryPolicy"]["maxRetries"] == 3
    assert described_http_route["retryPolicy"]["perRetryTimeout"]["unit"] == "s"
    assert described_http_route["retryPolicy"]["perRetryTimeout"]["value"] == 2
    assert described_http_route["retryPolicy"]["tcpRetryEvents"] == ["connection-reset"]

    assert described_http_route["timeout"]["idle"]["unit"] == "m"
    assert described_http_route["timeout"]["idle"]["value"] == 5
    assert described_http_route["timeout"]["perRequest"]["unit"] == "ms"
    assert described_http_route["timeout"]["perRequest"]["value"] == 500
    connection = client.delete_route(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        routeName=ROUTE_4,
        virtualRouterName=ROUTER_NAME,
    )

    route = connection["route"]
    assert route["status"]["status"] == "DELETED"

    spec = route["spec"]
    assert spec["priority"] == 4
    deleted_route = tcp_route_spec["tcpRoute"]

    assert len(deleted_route["action"]["weightedTargets"]) == 1
    weighted_target = deleted_route["action"]["weightedTargets"][0]
    assert weighted_target["port"] == 22
    assert weighted_target["virtualNode"] == "ssh-server-node"
    assert weighted_target["weight"] == 100

    assert deleted_route["match"]["port"] == 22

    assert deleted_route["timeout"]["idle"]["unit"] == "s"
    assert deleted_route["timeout"]["idle"]["value"] == 600

    with pytest.raises(ClientError) as e:
        client.describe_route(
            meshName=MESH_NAME,
            meshOwner=mesh_owner,
            routeName=ROUTE_4,
            virtualRouterName=ROUTER_NAME,
        )
    err = e.value.response["Error"]
    assert err["Code"] == "RouteNotFound"
    assert (
        err["Message"]
        == f"There is no route named {ROUTE_4} associated with router {ROUTER_NAME} in mesh {MESH_NAME}."
    )


@mock_aws
def test_list_routes_paginated(client):
    ROUTER_NAME = "mock_virtual_router"

    mesh = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )["mesh"]
    mesh_owner = mesh["metadata"]["meshOwner"]

    client.create_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_NAME,
        spec={"listeners": [{"portMapping": {"port": 80, "protocol": "http"}}]},
        tags=[{"key": "router_traffic", "value": "http"}],
    )

    for idx, license in enumerate(["apache", "mit", "mpl", "bsd"]):
        client.create_route(
            meshOwner=mesh_owner,
            meshName=MESH_NAME,
            virtualRouterName=ROUTER_NAME,
            routeName=f"route{idx}",
            tags=[{"key": "license", "value": license}],
            spec=grpc_route_spec,
        )

    all_routes = client.list_routes(meshName=MESH_NAME, virtualRouterName=ROUTER_NAME)[
        "routes"
    ]
    assert len(all_routes) == 4

    page1 = client.list_routes(
        meshName=MESH_NAME, virtualRouterName=ROUTER_NAME, limit=2
    )
    assert len(page1["routes"]) == 2

    page2 = client.list_routes(
        meshName=MESH_NAME, virtualRouterName=ROUTER_NAME, nextToken=page1["nextToken"]
    )
    assert len(page2["routes"]) == 2


@mock_aws
def test_create_describe_list_update_delete_virtual_node(client):
    connection = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    assert "mesh" in connection
    mesh = connection["mesh"]
    assert "metadata" in mesh
    mesh_owner = mesh["metadata"]["meshOwner"]

    GRPC_NODE = "grpc_node"
    HTTP_NODE = "http_node"
    HTTP2_NODE = "http2_node"
    TCP_NODE = "tcp_node"

    connection = client.create_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=grpc_virtual_node_spec,
        tags=[{"key": "type", "value": "grpc"}],
        virtualNodeName=GRPC_NODE,
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is True
    assert tls["ports"] == [443]
    assert tls["certificate"]["file"]["certificateChain"] == "/path/to/cert_chain.pem"
    assert tls["certificate"]["file"]["privateKey"] == "/path/to/private_key.pem"
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "grpc.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is True
    assert backend["virtualService"]["clientPolicy"]["tls"]["ports"] == [443]
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == [
        "validation-alternate-name"
    ]
    assert validation["trust"]["acm"]["certificateAuthorityArns"] == ["example-acm-arn"]
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "my-grpc-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["grpc"]["maxRequests"] == 500
    assert listener["healthCheck"]["healthyThreshold"] == 2
    assert listener["healthCheck"]["intervalMillis"] == 5000
    assert listener["healthCheck"]["port"] == 50051
    assert listener["healthCheck"]["protocol"] == "grpc"
    assert listener["healthCheck"]["timeoutMillis"] == 2000
    assert listener["healthCheck"]["unhealthyThreshold"] == 3
    assert listener["portMapping"]["port"] == 50051
    assert listener["portMapping"]["protocol"] == "grpc"
    assert listener["timeout"]["grpc"]["idle"]["unit"] == "s"
    assert listener["timeout"]["grpc"]["idle"]["value"] == 600
    assert listener["timeout"]["grpc"]["perRequest"]["unit"] == "s"
    assert listener["timeout"]["grpc"]["perRequest"]["value"] == 30
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
    )
    assert listener["tls"]["mode"] == "STRICT"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "my-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/new_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "end_time",
        "value": "%END_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "status_code",
        "value": "%RESPONSE_CODE%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {"key": "region", "value": "us-east-1"}
    assert service_discovery["ipPreference"] == "IPv6_PREFERRED"
    assert service_discovery["namespaceName"] == "new-namespace"
    assert service_discovery["serviceName"] == "new-service"

    connection = client.create_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=http_virtual_node_spec,
        tags=[{"key": "type", "value": "http"}],
        virtualNodeName=HTTP_NODE,
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]

    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is True
    assert tls["ports"] == [443]
    assert tls["certificate"]["file"]["certificateChain"] == "/path/to/cert_chain.pem"
    assert tls["certificate"]["file"]["privateKey"] == "/path/to/private_key.pem"
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "www.example.com",
        "api.example.com",
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is False
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == [
        "example-alternative-name"
    ]
    assert (
        validation["trust"]["file"]["certificateChain"] == "example-certificate-chain"
    )
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "my-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["http"]["maxConnections"] == 1000
    assert listener["connectionPool"]["http"]["maxPendingRequests"] == 5000
    assert listener["healthCheck"]["healthyThreshold"] == 2
    assert listener["healthCheck"]["intervalMillis"] == 5000
    assert listener["healthCheck"]["path"] == "/health"
    assert listener["healthCheck"]["port"] == 80
    assert listener["healthCheck"]["protocol"] == "http"
    assert listener["healthCheck"]["timeoutMillis"] == 2000
    assert listener["healthCheck"]["unhealthyThreshold"] == 3
    assert listener["outlierDetection"]["baseEjectionDuration"]["unit"] == "s"
    assert listener["outlierDetection"]["baseEjectionDuration"]["value"] == 30
    assert listener["outlierDetection"]["interval"]["unit"] == "s"
    assert listener["outlierDetection"]["interval"]["value"] == 10
    assert listener["outlierDetection"]["maxEjectionPercent"] == 10
    assert listener["outlierDetection"]["maxServerErrors"] == 5
    assert listener["portMapping"]["port"] == 80
    assert listener["portMapping"]["protocol"] == "http"
    assert listener["timeout"]["http"]["idle"]["unit"] == "s"
    assert listener["timeout"]["http"]["idle"]["value"] == 60
    assert listener["timeout"]["http"]["perRequest"]["unit"] == "s"
    assert listener["timeout"]["http"]["perRequest"]["value"] == 5
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
    )
    assert listener["tls"]["mode"] == "STRICT"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "my-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "start_time",
        "value": "%START_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "method",
        "value": "%REQ(:METHOD)%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {"key": "env", "value": "prod"}
    assert service_discovery["ipPreference"] == "IPv4_PREFERRED"
    assert service_discovery["namespaceName"] == "my-namespace"
    assert service_discovery["serviceName"] == "my-service"

    connection = client.create_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=http2_virtual_node_spec,
        tags=[{"key": "type", "value": "http2"}],
        virtualNodeName=HTTP2_NODE,
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is True
    assert tls["ports"] == [443]
    assert tls["certificate"]["file"]["certificateChain"] == "/path/to/cert_chain.pem"
    assert tls["certificate"]["file"]["privateKey"] == "/path/to/private_key.pem"
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "http2.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is True
    assert backend["virtualService"]["clientPolicy"]["tls"]["ports"] == [443]
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == ["match-me"]
    assert validation["trust"]["sds"]["secretName"] == "example-secret-name"
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "my-http2-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["http2"]["maxRequests"] == 1000
    assert listener["healthCheck"]["healthyThreshold"] == 2
    assert listener["healthCheck"]["intervalMillis"] == 5000
    assert listener["healthCheck"]["path"] == "/"
    assert listener["healthCheck"]["port"] == 443
    assert listener["healthCheck"]["protocol"] == "http2"
    assert listener["healthCheck"]["timeoutMillis"] == 2000
    assert listener["healthCheck"]["unhealthyThreshold"] == 3
    assert listener["portMapping"]["port"] == 443
    assert listener["portMapping"]["protocol"] == "http2"
    assert listener["timeout"]["http2"]["idle"]["unit"] == "s"
    assert listener["timeout"]["http2"]["idle"]["value"] == 120
    assert listener["timeout"]["http2"]["perRequest"]["unit"] == "s"
    assert listener["timeout"]["http2"]["perRequest"]["value"] == 10
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
    )
    assert listener["tls"]["mode"] == "STRICT"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "my-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/new_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "end_time",
        "value": "%END_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "status_code",
        "value": "%RESPONSE_CODE%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {"key": "region", "value": "us-east-1"}
    assert service_discovery["ipPreference"] == "IPv6_PREFERRED"
    assert service_discovery["namespaceName"] == "new-namespace"
    assert service_discovery["serviceName"] == "new-service"

    connection = client.create_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=tcp_virtual_node_spec,
        tags=[{"key": "type", "value": "tcp"}],
        virtualNodeName=TCP_NODE,
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is True
    assert tls["ports"] == [443]
    assert tls["certificate"]["file"]["certificateChain"] == "/path/to/cert_chain.pem"
    assert tls["certificate"]["file"]["privateKey"] == "/path/to/private_key.pem"
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "tcp.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is False
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == [
        "exact-match-example"
    ]
    assert validation["trust"]["file"]["certificateChain"] == "test-certificate-chain"
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "my-tcp-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["tcp"]["maxConnections"] == 2000
    assert listener["healthCheck"]["healthyThreshold"] == 2
    assert listener["healthCheck"]["intervalMillis"] == 10000
    assert listener["healthCheck"]["port"] == 8080
    assert listener["healthCheck"]["protocol"] == "tcp"
    assert listener["healthCheck"]["timeoutMillis"] == 5000
    assert listener["healthCheck"]["unhealthyThreshold"] == 3
    assert listener["portMapping"]["port"] == 8080
    assert listener["portMapping"]["protocol"] == "tcp"
    assert listener["timeout"]["tcp"]["idle"]["unit"] == "m"
    assert listener["timeout"]["tcp"]["idle"]["value"] == 30

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/new_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "end_time",
        "value": "%END_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "status_code",
        "value": "%RESPONSE_CODE%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {"key": "region", "value": "us-east-1"}
    assert service_discovery["ipPreference"] == "IPv6_PREFERRED"
    assert service_discovery["namespaceName"] == "new-namespace"
    assert service_discovery["serviceName"] == "new-service"

    connection = client.list_virtual_nodes(meshName=MESH_NAME, meshOwner=mesh_owner)
    virtual_nodes = connection.get("virtualNodes")
    assert isinstance(virtual_nodes, list)
    assert len(virtual_nodes) == 4
    names_counted = defaultdict(int)
    for node in virtual_nodes:
        node_name = node.get("virtualNodeName")
        if node_name:
            names_counted[node_name] += 1
    assert names_counted[GRPC_NODE] == 1
    assert names_counted[HTTP_NODE] == 1
    assert names_counted[HTTP2_NODE] == 1
    assert names_counted[TCP_NODE] == 1

    connection = client.update_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=modified_http2_virtual_node_spec,
        virtualNodeName=HTTP2_NODE,
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is False
    assert tls["ports"] == [8443]
    assert (
        tls["certificate"]["file"]["certificateChain"]
        == "/updated/path/to/cert_chain.pem"
    )
    assert (
        tls["certificate"]["file"]["privateKey"] == "/updated/path/to/private_key.pem"
    )
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "updated.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/updated/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is False
    assert backend["virtualService"]["clientPolicy"]["tls"]["ports"] == [8443]
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "updated-http2-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["http2"]["maxRequests"] == 500
    assert listener["healthCheck"]["healthyThreshold"] == 3
    assert listener["healthCheck"]["intervalMillis"] == 5100
    assert listener["healthCheck"]["path"] == "/health"
    assert listener["healthCheck"]["port"] == 8443
    assert listener["healthCheck"]["protocol"] == "http2"
    assert listener["healthCheck"]["timeoutMillis"] == 2200
    assert listener["healthCheck"]["unhealthyThreshold"] == 2
    assert listener["portMapping"]["port"] == 8443
    assert listener["portMapping"]["protocol"] == "http2"
    assert listener["timeout"]["http2"]["idle"]["unit"] == "m"
    assert listener["timeout"]["http2"]["idle"]["value"] == 5
    assert listener["timeout"]["http2"]["perRequest"]["unit"] == "m"
    assert listener["timeout"]["http2"]["perRequest"]["value"] == 1
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-west-2:987654321098:certificate/hgfedcba-4321-8765-ba09-fedc09876543"
    )
    assert listener["tls"]["mode"] == "PERMISSIVE"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "updated-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/updated_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "start_time",
        "value": "%START_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "method",
        "value": "%REQUEST_METHOD%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {
        "key": "environment",
        "value": "production",
    }
    assert service_discovery["ipPreference"] == "IPv4_PREFERRED"
    assert service_discovery["namespaceName"] == "updated-namespace"
    assert service_discovery["serviceName"] == "updated-service"

    connection = client.describe_virtual_node(
        meshName=MESH_NAME, meshOwner=mesh_owner, virtualNodeName=HTTP2_NODE
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is False
    assert tls["ports"] == [8443]
    assert (
        tls["certificate"]["file"]["certificateChain"]
        == "/updated/path/to/cert_chain.pem"
    )
    assert (
        tls["certificate"]["file"]["privateKey"] == "/updated/path/to/private_key.pem"
    )
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "updated.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/updated/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is False
    assert backend["virtualService"]["clientPolicy"]["tls"]["ports"] == [8443]
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == [
        "another-exact-match-example"
    ]
    assert (
        validation["trust"]["file"]["certificateChain"]
        == "different-test-certificate-chain"
    )
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "updated-http2-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["http2"]["maxRequests"] == 500
    assert listener["healthCheck"]["healthyThreshold"] == 3
    assert listener["healthCheck"]["intervalMillis"] == 5100
    assert listener["healthCheck"]["path"] == "/health"
    assert listener["healthCheck"]["port"] == 8443
    assert listener["healthCheck"]["protocol"] == "http2"
    assert listener["healthCheck"]["timeoutMillis"] == 2200
    assert listener["healthCheck"]["unhealthyThreshold"] == 2
    assert listener["portMapping"]["port"] == 8443
    assert listener["portMapping"]["protocol"] == "http2"
    assert listener["timeout"]["http2"]["idle"]["unit"] == "m"
    assert listener["timeout"]["http2"]["idle"]["value"] == 5
    assert listener["timeout"]["http2"]["perRequest"]["unit"] == "m"
    assert listener["timeout"]["http2"]["perRequest"]["value"] == 1
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-west-2:987654321098:certificate/hgfedcba-4321-8765-ba09-fedc09876543"
    )
    assert listener["tls"]["mode"] == "PERMISSIVE"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "updated-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/updated_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "start_time",
        "value": "%START_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "method",
        "value": "%REQUEST_METHOD%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {
        "key": "environment",
        "value": "production",
    }
    assert service_discovery["ipPreference"] == "IPv4_PREFERRED"
    assert service_discovery["namespaceName"] == "updated-namespace"
    assert service_discovery["serviceName"] == "updated-service"

    connection = client.delete_virtual_node(
        meshName=MESH_NAME, meshOwner=mesh_owner, virtualNodeName=GRPC_NODE
    )
    virtualNode = connection["virtualNode"]
    spec = virtualNode["spec"]
    backend_defaults = spec["backendDefaults"]
    tls = backend_defaults["clientPolicy"]["tls"]
    assert tls["enforce"] is True
    assert tls["ports"] == [443]
    assert tls["certificate"]["file"]["certificateChain"] == "/path/to/cert_chain.pem"
    assert tls["certificate"]["file"]["privateKey"] == "/path/to/private_key.pem"
    assert tls["validation"]["subjectAlternativeNames"]["match"]["exact"] == [
        "grpc.example.com"
    ]
    assert (
        tls["validation"]["trust"]["file"]["certificateChain"]
        == "/path/to/ca_bundle.pem"
    )

    assert len(spec["backends"]) == 1
    backend = spec["backends"][0]
    assert backend["virtualService"]["clientPolicy"]["tls"]["enforce"] is True
    assert backend["virtualService"]["clientPolicy"]["tls"]["ports"] == [443]
    validation = backend["virtualService"]["clientPolicy"]["tls"]["validation"]
    assert validation["subjectAlternativeNames"]["match"]["exact"] == [
        "validation-alternate-name"
    ]
    assert validation["trust"]["acm"]["certificateAuthorityArns"] == ["example-acm-arn"]
    assert (
        backend["virtualService"]["virtualServiceName"]
        == "my-grpc-service.default.svc.cluster.local"
    )

    assert len(spec["listeners"]) == 1
    listener = spec["listeners"][0]
    assert listener["connectionPool"]["grpc"]["maxRequests"] == 500
    assert listener["healthCheck"]["healthyThreshold"] == 2
    assert listener["healthCheck"]["intervalMillis"] == 5000
    assert listener["healthCheck"]["port"] == 50051
    assert listener["healthCheck"]["protocol"] == "grpc"
    assert listener["healthCheck"]["timeoutMillis"] == 2000
    assert listener["healthCheck"]["unhealthyThreshold"] == 3
    assert listener["portMapping"]["port"] == 50051
    assert listener["portMapping"]["protocol"] == "grpc"
    assert listener["timeout"]["grpc"]["idle"]["unit"] == "s"
    assert listener["timeout"]["grpc"]["idle"]["value"] == 600
    assert listener["timeout"]["grpc"]["perRequest"]["unit"] == "s"
    assert listener["timeout"]["grpc"]["perRequest"]["value"] == 30
    assert (
        listener["tls"]["certificate"]["acm"]["certificateArn"]
        == "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
    )
    assert listener["tls"]["mode"] == "STRICT"
    assert (
        listener["tls"]["validation"]["trust"]["sds"]["secretName"]
        == "my-ca-bundle-secret"
    )

    access_log = spec["logging"]["accessLog"]
    assert access_log["file"]["path"] == "/var/log/appmesh/new_access.log"
    assert len(access_log["file"]["format"]["json"]) == 2
    assert access_log["file"]["format"]["json"][0] == {
        "key": "end_time",
        "value": "%END_TIME%",
    }
    assert access_log["file"]["format"]["json"][1] == {
        "key": "status_code",
        "value": "%RESPONSE_CODE%",
    }

    service_discovery = spec["serviceDiscovery"]["awsCloudMap"]
    assert len(service_discovery["attributes"]) == 1
    assert service_discovery["attributes"][0] == {"key": "region", "value": "us-east-1"}
    assert service_discovery["ipPreference"] == "IPv6_PREFERRED"
    assert service_discovery["namespaceName"] == "new-namespace"
    assert service_discovery["serviceName"] == "new-service"

    with pytest.raises(ClientError) as e:
        client.describe_virtual_node(
            meshName=MESH_NAME,
            meshOwner=mesh_owner,
            virtualNodeName=GRPC_NODE,
        )
    err = e.value.response["Error"]
    assert err["Code"] == "VirtualNodeNotFound"
    assert (
        err["Message"]
        == f"{GRPC_NODE} is not a virtual node associated with mesh {MESH_NAME}"
    )


@mock_aws
def test_list_virtual_nodes_paginated(client):
    mesh = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )["mesh"]
    mesh_owner = mesh["metadata"]["meshOwner"]

    for i in range(5):
        client.create_virtual_node(
            meshName=MESH_NAME,
            meshOwner=mesh_owner,
            spec=grpc_virtual_node_spec,
            tags=[{"key": "type", "value": "grpc"}],
            virtualNodeName=f"node{i}",
        )

    all_nodes = client.list_virtual_nodes(meshName=MESH_NAME, meshOwner=mesh_owner)[
        "virtualNodes"
    ]
    assert len(all_nodes) == 5

    page1 = client.list_virtual_nodes(meshName=MESH_NAME, meshOwner=mesh_owner, limit=2)
    assert len(page1["virtualNodes"]) == 2

    page2 = client.list_virtual_nodes(
        meshName=MESH_NAME, meshOwner=mesh_owner, nextToken=page1["nextToken"]
    )
    assert len(page2["virtualNodes"]) == 3


@mock_aws
def test_tag_and_list_tags_for_resource(client):
    # create resources
    connection = client.create_mesh(
        meshName=MESH_NAME,
        spec={
            "egressFilter": {"type": "DROP_ALL"},
            "serviceDiscovery": {"ipPreference": "IPv4_ONLY"},
        },
        tags=[{"key": "owner", "value": "moto"}],
    )
    mesh = connection["mesh"]
    mesh_owner = mesh["metadata"]["meshOwner"]
    mesh_arn = mesh["metadata"]["arn"]

    ROUTER_NAME = "mock_router"

    connection = client.create_virtual_router(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        virtualRouterName=ROUTER_NAME,
        spec={"listeners": [{"portMapping": {"port": 80, "protocol": "http"}}]},
        tags=[{"key": "router_traffic", "value": "http"}],
    )

    virtual_router = connection.get("virtualRouter")
    virtual_router_arn = virtual_router["metadata"]["arn"]

    connection = client.create_route(
        meshOwner=mesh_owner,
        meshName=MESH_NAME,
        virtualRouterName=ROUTER_NAME,
        routeName="mock_http_route",
        tags=[{"key": "license", "value": "mit"}],
        spec=http_route_spec,
    )

    route = connection["route"]
    route_arn = route["metadata"]["arn"]

    connection = client.create_virtual_node(
        meshName=MESH_NAME,
        meshOwner=mesh_owner,
        spec=http_virtual_node_spec,
        tags=[{"key": "type", "value": "http"}],
        virtualNodeName="mock_http_node",
    )
    virtual_node = connection["virtualNode"]
    virtual_node_arn = virtual_node["metadata"]["arn"]

    # add tags and validate
    client.tag_resource(
        resourceArn=mesh_arn, tags=[{"key": "organization", "value": "moto"}]
    )
    connection = client.list_tags_for_resource(resourceArn=mesh_arn)
    tags = connection["tags"]
    assert tags[0] == {"key": "owner", "value": "moto"}
    assert tags[1] == {"key": "organization", "value": "moto"}

    client.tag_resource(
        resourceArn=virtual_router_arn,
        tags=[{"key": "organization", "value": "2moto2furious"}],
    )
    connection = client.list_tags_for_resource(resourceArn=virtual_router_arn)
    tags = connection["tags"]
    assert tags[0] == {"key": "router_traffic", "value": "http"}
    assert tags[1] == {"key": "organization", "value": "2moto2furious"}

    client.tag_resource(
        resourceArn=route_arn, tags=[{"key": "organization", "value": "motyo_drift"}]
    )
    connection = client.list_tags_for_resource(resourceArn=route_arn)
    tags = connection["tags"]
    assert tags[0] == {"key": "license", "value": "mit"}
    assert tags[1] == {"key": "organization", "value": "motyo_drift"}

    client.tag_resource(
        resourceArn=virtual_node_arn,
        tags=[{"key": "k2", "value": "v2"}, {"key": "k3", "value": "v3"}],
    )
    connection = client.list_tags_for_resource(resourceArn=virtual_node_arn)
    tags = connection["tags"]
    assert tags[0] == {"key": "type", "value": "http"}
    assert tags[1] == {"key": "k2", "value": "v2"}
    assert tags[2] == {"key": "k3", "value": "v3"}

    page1 = client.list_tags_for_resource(resourceArn=virtual_node_arn, limit=1)
    assert page1["tags"] == [{"key": "type", "value": "http"}]

    page2 = client.list_tags_for_resource(
        resourceArn=virtual_node_arn, nextToken=page1["nextToken"]
    )
    assert page2["tags"] == [{"key": "k2", "value": "v2"}, {"key": "k3", "value": "v3"}]



from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, List, Optional, Set, Tuple, TypedDict

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import (
    HistoryEventType,
    LoggingConfiguration,
    LogLevel,
    LongArn,
)
from moto.utilities.arns import (
    parse_arn,
)

LOG = logging.getLogger(__name__)

ExecutionEventLogDetails = dict

# The following event type sets are compiled according to AWS's
# log level definitions: https://docs.aws.amazon.com/step-functions/latest/dg/cloudwatch-log-level.html
_ERROR_LOG_EVENT_TYPES: Set[HistoryEventType] = {
    HistoryEventType.ExecutionAborted,
    HistoryEventType.ExecutionFailed,
    HistoryEventType.ExecutionTimedOut,
    HistoryEventType.FailStateEntered,
    HistoryEventType.LambdaFunctionFailed,
    HistoryEventType.LambdaFunctionScheduleFailed,
    HistoryEventType.LambdaFunctionStartFailed,
    HistoryEventType.LambdaFunctionTimedOut,
    HistoryEventType.MapStateAborted,
    HistoryEventType.MapStateFailed,
    HistoryEventType.MapIterationAborted,
    HistoryEventType.MapIterationFailed,
    HistoryEventType.MapRunAborted,
    HistoryEventType.MapRunFailed,
    HistoryEventType.ParallelStateAborted,
    HistoryEventType.ParallelStateFailed,
    HistoryEventType.TaskFailed,
    HistoryEventType.TaskStartFailed,
    HistoryEventType.TaskStateAborted,
    HistoryEventType.TaskSubmitFailed,
    HistoryEventType.TaskTimedOut,
    HistoryEventType.WaitStateAborted,
}
_FATAL_LOG_EVENT_TYPES: Set[HistoryEventType] = {
    HistoryEventType.ExecutionAborted,
    HistoryEventType.ExecutionFailed,
    HistoryEventType.ExecutionTimedOut,
}


# The LogStreamName used when creating the empty Log Stream when validating the logging configuration.
VALIDATION_LOG_STREAM_NAME: str = (
    "log_stream_created_by_aws_to_validate_log_delivery_subscriptions"
)


def is_logging_enabled_for(
    log_level: LogLevel, history_event_type: HistoryEventType
) -> bool:
    # Checks whether the history event type is in the context of a give LogLevel.
    if log_level == LogLevel.ALL:
        return True
    elif log_level == LogLevel.OFF:
        return False
    elif log_level == LogLevel.ERROR:
        return history_event_type in _ERROR_LOG_EVENT_TYPES
    elif log_level == LogLevel.FATAL:
        return history_event_type in _FATAL_LOG_EVENT_TYPES
    else:
        LOG.error("Unknown LogLevel '%s'", log_level)


class CloudWatchLoggingConfiguration:
    state_machine_arn: LongArn
    log_level: LogLevel
    log_account_id: str
    log_region: str
    log_group_name: str
    log_stream_name: str
    include_execution_data: bool

    def __init__(
        self,
        state_machine_arn: LongArn,
        log_account_id: str,
        log_region: str,
        log_group_name: str,
        log_level: LogLevel,
        include_execution_data: bool,
    ):
        self.state_machine_arn = state_machine_arn
        self.log_level = log_level
        self.log_group_name = log_group_name
        self.log_account_id = log_account_id
        self.log_region = log_region
        # TODO: AWS appears to append a date and a serial number to the log
        #  stream name: more investigations are needed in this area.
        self.log_stream_name = f"states/{state_machine_arn}"
        self.include_execution_data = include_execution_data

    @staticmethod
    def extract_log_arn_parts_from(
        logging_configuration: LoggingConfiguration,
    ) -> Optional[Tuple[str, str, str]]:
        # Returns a tuple with: account_id, region, and log group name if the logging configuration
        # specifies a valid cloud watch log group arn, none otherwise.

        destinations = logging_configuration.get("destinations")
        if (
            not destinations or len(destinations) > 1
        ):  # Only one destination can be defined.
            return None

        log_group = destinations[0].get("cloudWatchLogsLogGroup")
        if not log_group:
            return None

        log_group_arn = log_group.get("logGroupArn")
        if not log_group_arn:
            return None

        try:
            arn_data = parse_arn(log_group_arn)
        except IndexError:
            return None

        log_region = arn_data.region
        if log_region is None:
            return None

        log_account_id = arn_data.account
        if log_account_id is None:
            return None

        log_resource = arn_data.resource_id
        if log_resource is None:
            return None

        log_resource_parts = log_resource.split("log-group:")
        if not log_resource_parts:
            return None

        log_group_name = log_resource_parts[-1].split(":")[0]
        return log_account_id, log_region, log_group_name

    @staticmethod
    def from_logging_configuration(
        state_machine_arn: LongArn,
        logging_configuration: LoggingConfiguration,
    ) -> Optional[CloudWatchLoggingConfiguration]:
        log_level = logging_configuration.get("level", LogLevel.OFF)
        if log_level == LogLevel.OFF:
            return None

        log_arn_parts = CloudWatchLoggingConfiguration.extract_log_arn_parts_from(
            logging_configuration=logging_configuration
        )
        if not log_arn_parts:
            return None
        log_account_id, log_region, log_group_name = log_arn_parts

        include_execution_data = logging_configuration["includeExecutionData"]

        return CloudWatchLoggingConfiguration(
            state_machine_arn=state_machine_arn,
            log_account_id=log_account_id,
            log_region=log_region,
            log_group_name=log_group_name,
            log_level=log_level,
            include_execution_data=include_execution_data,
        )

    def validate(self) -> None:
        pass


class HistoryLog(TypedDict):
    id: str
    previous_event_id: str
    event_timestamp: datetime
    type: HistoryEventType
    execution_arn: LongArn
    details: Optional[ExecutionEventLogDetails]


class CloudWatchLoggingSession:
    execution_arn: LongArn
    configuration: CloudWatchLoggingConfiguration
    _logs_client: BaseClient
    _is_log_stream_available: bool

    def __init__(
        self, execution_arn: LongArn, configuration: CloudWatchLoggingConfiguration
    ):
        self.execution_arn = execution_arn
        self.configuration = configuration
        self._logs_client = None

    def log_level_filter(self, history_event_type: HistoryEventType) -> bool:
        # Checks whether the history event type should be logged in this session.
        return is_logging_enabled_for(
            log_level=self.configuration.log_level,
            history_event_type=history_event_type,
        )

    def publish_history_log(self, history_log: HistoryLog) -> None:
        pass

    def _publish_history_log_or_setup(self, log_events: List[Any]):
        # Attempts to put the events into the given log group and stream, and attempts to create the stream if
        # this does not already exist.
        is_events_put = self._put_events(log_events=log_events)
        if is_events_put:
            return

        is_setup = self._setup()
        if not is_setup:
            LOG.debug(
                "CloudWatch Log was not published due to setup errors encountered "
                "while creating the LogStream for execution '%s'.",
                self.execution_arn,
            )
            return

        self._put_events(log_events=log_events)

    def _put_events(self, log_events: List[Any]) -> bool:
        # Puts the events to the targe log group and stream, and returns false if the LogGroup or LogStream could
        # not be found, true otherwise.
        try:
            self._logs_client.put_log_events(
                logGroupName=self.configuration.log_group_name,
                logStreamName=self.configuration.log_stream_name,
                logEvents=log_events,
            )
        except ClientError as error:
            error_code = error.response["Error"]["Code"]
            if error_code == "ResourceNotFoundException":
                return False
        except Exception as ignored:
            LOG.warning(
                "State Machine execution log event could not be published due to an error: '%s'",
                ignored,
            )
        return True

    def _setup(self) -> bool:
        # Create the log stream if one does not exist already.
        # TODO: enhance the verification logic to match AWS's logic to ensure IAM features work as expected.
        #  https://docs.aws.amazon.com/step-functions/latest/dg/cw-logs.html#cloudwatch-iam-policy
        try:
            self._logs_client.create_log_stream(
                logGroupName=self.configuration.log_group_name,
                logStreamName=self.configuration.log_stream_name,
            )
        except ClientError as error:
            error_code = error.response["Error"]["Code"]
            if error_code != "ResourceAlreadyExistsException":
                LOG.error(
                    "Could not create execution log stream for execution '%s' due to %s",
                    self.execution_arn,
                    error,
                )
                return False
        return True



"""Unit tests for directconnect-supported APIs."""

import time

import boto3
import pytest

from moto import mock_aws


@pytest.fixture(name="client")
def fixture_dx_client():
    with mock_aws():
        yield boto3.client("directconnect", region_name="us-east-1")


def test_create_connection(client):
    connection = client.create_connection(
        location="EqDC2", bandwidth="10Gbps", connectionName="TestConnection"
    )
    assert connection["connectionId"].startswith("dx-moto")
    assert connection["connectionState"] == "available"


def test_describe_connections(client):
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
        requestMACSec=False,
    )
    time.sleep(0.1)
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection2",
        requestMACSec=True,
    )
    time.sleep(0.1)
    resp = client.describe_connections()
    connections = resp["connections"]
    assert len(connections) == 2
    assert not connections[0]["macSecCapable"]
    assert connections[1]["macSecCapable"]
    assert len(connections[0]["macSecKeys"]) == 0
    assert len(connections[1]["macSecKeys"]) == 1
    assert connections[0]["encryptionMode"] == "no_encrypt"
    assert connections[1]["encryptionMode"] == "must_encrypt"
    resp = client.describe_connections(connectionId=connections[0]["connectionId"])
    assert len(resp["connections"]) == 1


def test_delete_connection(client):
    connection = client.create_connection(
        location="EqDC2", bandwidth="10Gbps", connectionName="TestConnection"
    )
    connection = client.delete_connection(connectionId=connection["connectionId"])
    assert connection["connectionState"] == "deleted"


def test_update_connection(client):
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
    )
    time.sleep(0.1)
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection2",
        requestMACSec=True,
    )
    time.sleep(0.1)
    resp = client.describe_connections()
    connection1_id = resp["connections"][0]["connectionId"]
    connection2_id = resp["connections"][1]["connectionId"]
    connection = client.update_connection(
        connectionId=connection1_id,
        connectionName="NewConnectionName",
    )
    assert connection["connectionName"] == "NewConnectionName"
    connection = client.update_connection(
        connectionId=connection2_id,
        encryptionMode="should_encrypt",
    )
    assert connection["encryptionMode"] == "should_encrypt"


def test_associate_mac_sec_key_connection(client):
    connection = client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
    )
    resp = client.associate_mac_sec_key(
        connectionId=connection["connectionId"], ckn="_fake_ckn_", cak="_fake_cak_"
    )
    assert resp["connectionId"] == connection["connectionId"]
    mac_sec_keys = resp["macSecKeys"]
    assert len(mac_sec_keys) == 1
    assert mac_sec_keys[0]["ckn"] == "_fake_ckn_"
    assert "cak" not in mac_sec_keys[0]
    assert mac_sec_keys[0]["secretARN"] == "mock_secret_arn"


def test_associate_mac_sec_key_lag(client):
    lag = client.create_lag(
        numberOfConnections=1,
        location="eqDC2",
        connectionsBandwidth="10Gbps",
        lagName="TestLag1",
    )
    resp = client.associate_mac_sec_key(
        connectionId=lag["lagId"], ckn="_fake_ckn_", cak="_fake_cak_"
    )
    assert resp["connectionId"] == lag["lagId"]
    mac_sec_keys = resp["macSecKeys"]
    assert len(mac_sec_keys) == 1
    assert mac_sec_keys[0]["ckn"] == "_fake_ckn_"
    assert "cak" not in mac_sec_keys[0]
    assert mac_sec_keys[0]["secretARN"] == "mock_secret_arn"


def test_create_lag(client):
    lag = client.create_lag(
        numberOfConnections=1,
        location="eqDC2",
        connectionsBandwidth="10Gbps",
        lagName="TestLag0",
    )
    assert lag["lagId"].startswith("dxlag-moto")
    assert lag["lagState"] == "available"
    assert len(lag["connections"]) == 1
    connection = lag["connections"][0]
    assert connection["connectionName"].startswith(
        "Requested Connection 1 for Lag dxlag-moto"
    )


def test_describe_lags(client):
    client.create_lag(
        location="EqDC2",
        connectionsBandwidth="10Gbps",
        lagName="TestLag1",
        numberOfConnections=1,
        requestMACSec=False,
    )
    time.sleep(0.1)
    client.create_lag(
        location="EqDC2",
        connectionsBandwidth="10Gbps",
        lagName="TestLag2",
        numberOfConnections=1,
        requestMACSec=True,
    )
    time.sleep(0.1)
    resp = client.describe_lags()
    lags = resp["lags"]
    assert len(lags) == 2
    assert len(lags[0]["connections"]) == 1
    assert len(lags[1]["connections"]) == 1
    assert not lags[0]["macSecCapable"]
    assert not lags[0]["connections"][0]["macSecCapable"]
    assert lags[1]["macSecCapable"]
    assert lags[1]["connections"][0]["macSecCapable"]
    assert len(lags[0]["macSecKeys"]) == 0
    assert len(lags[0]["connections"][0]["macSecKeys"]) == 0
    assert len(lags[1]["macSecKeys"]) == 1
    assert len(lags[1]["connections"][0]["macSecKeys"]) == 1
    assert lags[0]["encryptionMode"] == "no_encrypt"
    assert lags[0]["connections"][0]["encryptionMode"] == "no_encrypt"
    assert lags[1]["encryptionMode"] == "must_encrypt"
    assert lags[1]["connections"][0]["encryptionMode"] == "must_encrypt"
    resp = client.describe_lags(lagId=lags[0]["lagId"])
    assert len(resp["lags"]) == 1


def _test_disassociate_mac_sec_key_common(client, connection_id: str):
    secret_arn = "_fake_secret_arn_"
    assoc_resp = client.associate_mac_sec_key(
        connectionId=connection_id,
        ckn="_fake_ckn_",
        cak="_fake_cak_",
        secretARN=secret_arn,
    )
    assert assoc_resp["connectionId"] == connection_id
    resp = client.disassociate_mac_sec_key(
        connectionId=connection_id, secretARN=secret_arn
    )
    assert resp["connectionId"] == connection_id
    mac_sec_keys = resp["macSecKeys"]
    assert len(mac_sec_keys) == 1
    assert mac_sec_keys[0]["secretARN"] == secret_arn
    assert mac_sec_keys[0]["state"] == "disassociated"


def test_disassociate_mac_sec_key_connection(client):
    connection = client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
    )
    connection_id = connection["connectionId"]
    _test_disassociate_mac_sec_key_common(client=client, connection_id=connection_id)


def test_disassociate_mac_sec_key_lag(client):
    lag = client.create_lag(
        numberOfConnections=1,
        location="eqDC2",
        connectionsBandwidth="10Gbps",
        lagName="TestLag1",
    )
    lag_id = lag["lagId"]
    _test_disassociate_mac_sec_key_common(client=client, connection_id=lag_id)



from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_component import (
    IterationComponent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.distributed_iterator import (
    DistributedIterator,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.inline_iterator import (
    InlineIterator,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_decl import (
    IteratorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.mode import (
    Mode,
)


def from_iterator_decl(iterator_decl: IteratorDecl) -> IterationComponent:
    if iterator_decl.processor_config.mode == Mode.Inline:
        return InlineIterator(
            query_language=iterator_decl.query_language,
            start_at=iterator_decl.start_at,
            states=iterator_decl.states,
            comment=iterator_decl.comment,
            processor_config=iterator_decl.processor_config,
        )
    elif iterator_decl.processor_config.mode == Mode.Distributed:
        return DistributedIterator(
            query_language=iterator_decl.query_language,
            start_at=iterator_decl.start_at,
            states=iterator_decl.states,
            comment=iterator_decl.comment,
            processor_config=iterator_decl.processor_config,
        )
    else:
        raise ValueError(
            f"Unknown Map state processing mode: '{iterator_decl.processor_config.mode}'."
        )



from typing import Any, Set

from moto.stepfunctions.parser.asl.component.common.flow.end import End
from moto.stepfunctions.parser.asl.component.common.flow.next import Next
from moto.stepfunctions.parser.asl.component.common.parargs import Parargs
from moto.stepfunctions.parser.asl.component.common.path.input_path import InputPath
from moto.stepfunctions.parser.asl.component.common.path.items_path import ItemsPath
from moto.stepfunctions.parser.asl.component.common.path.output_path import OutputPath
from moto.stepfunctions.parser.asl.component.common.timeouts.heartbeat import Heartbeat
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import Timeout
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_type import (
    Comparison,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.variable import (
    Variable,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.max_items_decl import (
    MaxItemsDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.items.items import (
    Items,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.max_concurrency import (
    MaxConcurrencyDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.tolerated_failure import (
    ToleratedFailureCountDecl,
    ToleratedFailurePercentageDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    Resource,
)
from moto.stepfunctions.parser.asl.component.state.fail.cause_decl import CauseDecl
from moto.stepfunctions.parser.asl.component.state.fail.error_decl import ErrorDecl
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps

UNIQUE_SUBINSTANCES: Set[type] = {
    InputPath,
    Items,
    ItemsPath,
    OutputPath,
    Resource,
    WaitFunction,
    Timeout,
    Heartbeat,
    MaxItemsDecl,
    MaxConcurrencyDecl,
    ToleratedFailureCountDecl,
    ToleratedFailurePercentageDecl,
    ErrorDecl,
    CauseDecl,
    Variable,
    Parargs,
    Comparison,
}


class StateProps(TypedProps):
    name: str

    def add(self, instance: Any) -> None:
        inst_type = type(instance)

        # End-Next conflicts:
        if inst_type == End and Next in self._instance_by_type:
            raise ValueError(
                f"End redefines Next, from '{self.get(Next)}' to '{instance}'."
            )
        if inst_type == Next and End in self._instance_by_type:
            raise ValueError(
                f"Next redefines End, from '{self.get(End)}' to '{instance}'."
            )

        # Subclasses
        for typ in UNIQUE_SUBINSTANCES:
            if issubclass(inst_type, typ):
                super()._add(typ, instance)
                return

        # Base and delegate to preprocessor.
        super().add(instance)



from moto.stepfunctions.parser.asl.eval.program_state import ProgramState


class ProgramChoiceSelected(ProgramState):
    next_state_name: str

    def __init__(self, next_state_name: str):
        super().__init__()
        self.next_state_name = next_state_name



from .responses import ElasticTranscoderResponse

url_bases = [
    r"https?://elastictranscoder\.(.+)\.amazonaws.com",
]


url_paths = {
    r"{0}/(?P<api_version>[^/]+)/pipelines/?$": ElasticTranscoderResponse.dispatch,
    r"{0}/(?P<api_version>[^/]+)/pipelines/(?P<pipeline_id>[^/]+)/?$": ElasticTranscoderResponse.dispatch,
}



import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_filter():
    client = boto3.client("inspector2", region_name="us-east-2")
    resp = client.create_filter(
        name="my_first_filter",
        reason="because I said so",
        action="NONE",
        description="my filter",
        filterCriteria={
            "codeVulnerabilityDetectorName": [{"comparison": "EQUALS", "value": "cvdn"}]
        },
    )
    assert "arn" in resp


@mock_aws
def test_list_filters():
    client = boto3.client("inspector2", region_name="ap-southeast-1")
    assert client.list_filters()["filters"] == []

    arn1 = client.create_filter(
        name="my_first_filter",
        action="NONE",
        filterCriteria={"findingArn": [{"comparison": "EQUALS", "value": "cvdn"}]},
    )["arn"]

    filters = client.list_filters()["filters"]
    assert len(filters) == 1
    assert filters[0]["arn"] == arn1

    arn2 = client.create_filter(
        name="my_second_filter",
        action="SUPPRESS",
        filterCriteria={"fixAvailable": [{"comparison": "EQUALS", "value": "cvdn"}]},
    )["arn"]

    filters = client.list_filters()["filters"]
    assert len(filters) == 2

    filters = client.list_filters(action="SUPPRESS")["filters"]
    assert len(filters) == 1
    assert filters[0]["arn"] == arn2

    filters = client.list_filters(arns=[arn1])["filters"]
    assert len(filters) == 1
    assert filters[0]["arn"] == arn1

    client.delete_filter(arn=arn1)

    filters = client.list_filters()["filters"]
    assert len(filters) == 1
    assert filters[0]["arn"] == arn2



from typing import Any, Dict, List, Optional

from moto.appmesh.dataclasses.route import (
    GrcpRouteRetryPolicy,
    GrpcMetadatum,
    GrpcRoute,
    GrpcRouteMatch,
    HttpRoute,
    HttpRouteMatch,
    HttpRouteRetryPolicy,
    Match,
    QueryParameterMatch,
    Range,
    RouteAction,
    RouteActionWeightedTarget,
    RouteMatchPath,
    RouteMatchQueryParameter,
    RouteSpec,
    TCPRoute,
    TCPRouteMatch,
)
from moto.appmesh.dataclasses.shared import Duration, Timeout
from moto.appmesh.dataclasses.virtual_node import (
    ACM,
    DNS,
    SDS,
    AccessLog,
    AccessLogFile,
    AWSCloudMap,
    Backend,
    BackendDefaults,
    BackendTrust,
    Certificate,
    CertificateFile,
    CertificateFileWithPrivateKey,
    ClientPolicy,
    ConnectionPool,
    GRPCOrHTTP2Connection,
    HealthCheck,
    HTTPConnection,
    KeyValue,
    Listener,
    ListenerCertificateACM,
    ListenerTLS,
    Logging,
    LoggingFormat,
    OutlierDetection,
    PortMapping,
    ProtocolTimeouts,
    ServiceDiscovery,
    SubjectAlternativeNames,
    TCPConnection,
    TCPTimeout,
    TLSBackendValidation,
    TLSClientPolicy,
    TLSListenerCertificate,
    TLSListenerValidation,
    Trust,
    VirtualNodeSpec,
    VirtualService,
)
from moto.appmesh.dataclasses.virtual_node import (
    Match as VirtualNodeMatch,
)
from moto.appmesh.dataclasses.virtual_router import PortMapping as RouterPortMapping
from moto.appmesh.exceptions import (
    MissingRequiredFieldError,
)


def port_mappings_from_router_spec(spec: Any) -> List[RouterPortMapping]:  # type: ignore[misc]
    return [
        RouterPortMapping(
            port=(listener.get("portMapping") or {}).get("port"),
            protocol=(listener.get("portMapping") or {}).get("protocol"),
        )
        for listener in ((spec or {}).get("listeners") or [])
    ]


def get_action_from_route(route: Any) -> RouteAction:  # type: ignore[misc]
    weighted_targets = [
        RouteActionWeightedTarget(
            port=target.get("port"),
            virtual_node=target.get("virtualNode"),
            weight=target.get("weight"),
        )
        for target in (route.get("action") or {}).get("weightedTargets") or []
    ]
    return RouteAction(weighted_targets=weighted_targets)


def get_route_match_metadata(metadata: List[Any]) -> List[GrpcMetadatum]:  # type: ignore[misc]
    output = []
    for _metadatum in metadata:
        _match = _metadatum.get("match")
        match = None
        if _match is not None:
            _range = _match.get("range")
            range = None
            if _range is not None:
                range = Range(start=_range.get("start"), end=_range.get("end"))
            match = Match(
                exact=_match.get("exact"),
                prefix=_match.get("prefix"),
                range=range,
                regex=_match.get("regex"),
                suffix=_match.get("suffix"),
            )
        output.append(
            GrpcMetadatum(
                invert=_metadatum.get("invert"),
                match=match,
                name=_metadatum.get("name"),
            )
        )
    return output


def get_grpc_route_match(route: Any) -> GrpcRouteMatch:  # type: ignore[misc]
    _route_match = route.get("match")
    metadata = None
    if _route_match is not None:
        metadata = get_route_match_metadata(_route_match.get("metadata") or [])
    return GrpcRouteMatch(
        metadata=metadata,
        method_name=_route_match.get("methodName"),
        port=_route_match.get("port"),
        service_name=_route_match.get("serviceName"),
    )


def get_http_match_from_route(route: Any) -> HttpRouteMatch:  # type: ignore[misc]
    _route_match = route.get("match") or {}
    headers, path, query_parameters = None, None, None
    if _route_match is not None:
        headers = get_route_match_metadata(_route_match.get("headers") or [])
        _path = _route_match.get("path")
        if _path is not None:
            path = RouteMatchPath(exact=_path.get("exact"), regex=_path.get("regex"))
        _query_parameters = _route_match.get("queryParameters")
        if _query_parameters is not None:
            query_parameters = []
            for _param in _query_parameters:
                _match = _param.get("match")
                match = None
                if _match is not None:
                    match = QueryParameterMatch(exact=_match.get("exact"))
                query_parameters.append(
                    RouteMatchQueryParameter(name=_param.get("name"), match=match)
                )
    return HttpRouteMatch(
        headers=headers,
        method=(_route_match or {}).get("method"),
        path=path,
        port=(_route_match or {}).get("port"),
        prefix=(_route_match or {}).get("prefix"),
        query_parameters=query_parameters,
        scheme=(_route_match or {}).get("scheme"),
    )


def get_http_retry_policy_from_route(route: Any) -> Optional[HttpRouteRetryPolicy]:  # type: ignore[misc]
    _retry_policy = route.get("retryPolicy")
    retry_policy = None
    if _retry_policy is not None:
        _per_retry_timeout = _retry_policy.get("perRetryTimeout")
        per_retry_timeout = Duration(
            unit=_per_retry_timeout.get("unit"), value=_per_retry_timeout.get("value")
        )
        retry_policy = HttpRouteRetryPolicy(
            max_retries=_retry_policy.get("maxRetries"),
            http_retry_events=_retry_policy.get("httpRetryEvents"),
            per_retry_timeout=per_retry_timeout,
            tcp_retry_events=_retry_policy.get("tcpRetryEvents"),
        )
    return retry_policy


def get_timeout_from_route(route: Any) -> Optional[Timeout]:  # type: ignore[misc]
    _timeout = route.get("timeout") or {}
    idle, per_request = None, None
    _idle = _timeout.get("idle")
    if _idle is not None:
        idle = Duration(unit=_idle.get("unit"), value=_idle.get("value"))
    _per_request = _timeout.get("perRequest")
    if _per_request is not None:
        per_request = Duration(
            unit=_per_request.get("unit"), value=_per_request.get("value")
        )
    return (
        Timeout(idle=idle, per_request=per_request)
        if idle is not None or per_request is not None
        else None
    )


def get_tls_for_client_policy(tls: Any) -> TLSClientPolicy:  # type: ignore[misc]
    _certificate = tls.get("certificate")
    _validation = tls.get("validation")
    certificate, validation = None, None
    if _certificate is not None:
        _file = _certificate.get("file")
        _sds = _certificate.get("sds")
        file, sds = None, None
        if _file is not None:
            file = CertificateFileWithPrivateKey(
                certificate_chain=_file.get("certificateChain"),
                private_key=_file.get("privateKey"),
            )
        if _sds is not None:
            sds = SDS(secret_name=_sds.get("secretName"))
        certificate = Certificate(file=file, sds=sds)
    if _validation is None:
        raise MissingRequiredFieldError("validation")
    _subject_alternative_names = _validation.get("subjectAlternativeNames")
    _trust = _validation.get("trust")
    subject_alternative_names = None

    if _subject_alternative_names is not None:
        match = VirtualNodeMatch(
            exact=(_subject_alternative_names.get("match") or {}).get("exact") or []
        )
        subject_alternative_names = SubjectAlternativeNames(match=match)

    if _trust is None:
        raise MissingRequiredFieldError("trust")

    _trust_file = _trust.get("file")
    _trust_sds = _trust.get("sds")
    _acm = _trust.get("acm")
    trust_file, trust_sds, acm = None, None, None
    if _trust_file is not None:
        trust_file = CertificateFile(
            certificate_chain=_trust_file.get("certificateChain")
        )
    if _trust_sds is not None:
        trust_sds = SDS(secret_name=_trust_sds.get("secretName"))
    if _acm is not None:
        acm = ACM(certificate_authority_arns=_acm.get("certificateAuthorityArns"))
    trust = BackendTrust(file=trust_file, sds=trust_sds, acm=acm)

    validation = TLSBackendValidation(
        subject_alternative_names=subject_alternative_names, trust=trust
    )
    return TLSClientPolicy(
        certificate=certificate,
        enforce=tls.get("enforce"),
        ports=tls.get("ports"),
        validation=validation,
    )


def build_route_spec(spec: Dict[str, Any]) -> RouteSpec:  # type: ignore[misc]
    _grpc_route = spec.get("grpcRoute")
    _http_route = spec.get("httpRoute")
    _http2_route = spec.get("http2Route")
    _tcp_route = spec.get("tcpRoute")
    grpc_route, http_route, http2_route, tcp_route = None, None, None, None
    if _grpc_route is not None:
        grpc_action = get_action_from_route(_grpc_route)
        grpc_route_match = get_grpc_route_match(_grpc_route)

        _retry_policy = _grpc_route.get("retryPolicy")
        grpc_retry_policy = None
        if _retry_policy is not None:
            _per_retry_timeout = _retry_policy.get("perRetryTimeout")
            per_retry_timeout = Duration(
                unit=_per_retry_timeout.get("unit"),
                value=_per_retry_timeout.get("value"),
            )
            grpc_retry_policy = GrcpRouteRetryPolicy(
                grpc_retry_events=_retry_policy.get("grpcRetryEvents"),
                http_retry_events=_retry_policy.get("httpRetryEvents"),
                max_retries=_retry_policy.get("maxRetries"),
                per_retry_timeout=per_retry_timeout,
                tcp_retry_events=_retry_policy.get("tcpRetryEvents"),
            )

        grpc_timeout = get_timeout_from_route(_grpc_route)

        grpc_route = GrpcRoute(
            action=grpc_action,
            match=grpc_route_match,
            retry_policy=grpc_retry_policy,
            timeout=grpc_timeout,
        )

    if _http_route is not None:
        http_action = get_action_from_route(_http_route)
        http_match = get_http_match_from_route(_http_route)
        http_retry_policy = get_http_retry_policy_from_route(_http_route)
        http_timeout = get_timeout_from_route(_http_route)

        http_route = HttpRoute(
            action=http_action,
            match=http_match,
            retry_policy=http_retry_policy,
            timeout=http_timeout,
        )

    if _http2_route is not None:
        http2_action = get_action_from_route(_http2_route)
        http2_match = get_http_match_from_route(_http2_route)
        http2_retry_policy = get_http_retry_policy_from_route(_http2_route)
        http2_timeout = get_timeout_from_route(_http2_route)

        http2_route = HttpRoute(
            action=http2_action,
            match=http2_match,
            retry_policy=http2_retry_policy,
            timeout=http2_timeout,
        )

    if _tcp_route is not None:
        tcp_action = get_action_from_route(_tcp_route)
        tcp_timeout = get_timeout_from_route(_tcp_route)

        _tcp_match = _tcp_route.get("match")
        tcp_match = None
        if _tcp_match is not None:
            tcp_match = TCPRouteMatch(port=_tcp_match.get("port"))

        tcp_route = TCPRoute(action=tcp_action, match=tcp_match, timeout=tcp_timeout)

    return RouteSpec(
        grpc_route=grpc_route,
        http_route=http_route,
        http2_route=http2_route,
        priority=spec.get("priority"),
        tcp_route=tcp_route,
    )


def build_virtual_node_spec(spec: Dict[str, Any]) -> VirtualNodeSpec:  # type: ignore[misc]
    _backend_defaults = spec.get("backendDefaults")
    _backends = spec.get("backends")
    _listeners = spec.get("listeners")
    _logging = spec.get("logging")
    _service_discovery = spec.get("serviceDiscovery")

    backend_defaults, backends, listeners, logging, service_discovery = (
        None,
        None,
        None,
        None,
        None,
    )

    if _backend_defaults is not None:
        _client_policy = _backend_defaults.get("clientPolicy")
        client_policy = None
        if _client_policy is not None:
            _tls = _client_policy.get("tls")
            tls = None
            if _tls is not None:
                tls = get_tls_for_client_policy(_tls)
            client_policy = ClientPolicy(tls=tls)

        backend_defaults = BackendDefaults(client_policy=client_policy)

    if _backends is not None:
        backends = []
        for _backend in _backends:
            _virtual_service = _backend.get("virtualService")
            virtual_service = None
            if _virtual_service is not None:
                _virtual_service_client_policy = _virtual_service.get("clientPolicy")
                virtual_service_client_policy = None
                if _virtual_service_client_policy is not None:
                    _tls_client_policy = _virtual_service_client_policy.get("tls")
                    tls_client_policy = None
                    if _tls_client_policy is not None:
                        tls_client_policy = get_tls_for_client_policy(
                            _tls_client_policy
                        )
                    virtual_service_client_policy = ClientPolicy(tls=tls_client_policy)
                virtual_service = VirtualService(
                    client_policy=virtual_service_client_policy,
                    virtual_service_name=_virtual_service.get("virtualServiceName"),
                )
            backend = Backend(virtual_service=virtual_service)
            backends.append(backend)

    if _listeners is not None:
        listeners = []
        for _listener in _listeners:
            _connection_pool = _listener.get("connectionPool")
            _health_check = _listener.get("healthCheck")
            _outlier_detection = _listener.get("outlierDetection")
            _port_mapping = _listener.get("portMapping")
            _timeout = _listener.get("timeout")
            _listener_tls = _listener.get("tls")
            (
                connection_pool,
                health_check,
                outlier_detection,
                timeout,
                listener_tls,
            ) = None, None, None, None, None

            if _connection_pool is not None:
                _grpc = _connection_pool.get("grpc")
                _http = _connection_pool.get("http")
                _http2 = _connection_pool.get("http2")
                _tcp = _connection_pool.get("tcp")
                grpc, http, http2, tcp = None, None, None, None

                if _grpc is not None:
                    grpc = GRPCOrHTTP2Connection(max_requests=_grpc.get("maxRequests"))
                if _http is not None:
                    http = HTTPConnection(
                        max_connections=_http.get("maxConnections"),
                        max_pending_requests=_http.get("maxPendingRequests"),
                    )
                if _http2 is not None:
                    http2 = GRPCOrHTTP2Connection(
                        max_requests=_http2.get("maxRequests")
                    )
                if _tcp is not None:
                    tcp = TCPConnection(max_connections=_tcp.get("maxConnections"))

                connection_pool = ConnectionPool(
                    grpc=grpc, http=http, http2=http2, tcp=tcp
                )
            if _health_check is not None:
                health_check = HealthCheck(
                    healthy_threshold=_health_check.get("healthyThreshold"),
                    interval_millis=_health_check.get("intervalMillis"),
                    path=_health_check.get("path"),
                    port=_health_check.get("port"),
                    protocol=_health_check.get("protocol"),
                    timeout_millis=_health_check.get("timeoutMillis"),
                    unhealthy_threshold=_health_check.get("unhealthyThreshold"),
                )

            if _outlier_detection is not None:
                _base_ejection_duration = _outlier_detection.get("baseEjectionDuration")
                _interval = _outlier_detection.get("interval")
                if _base_ejection_duration is None:
                    raise MissingRequiredFieldError("baseEjectionDuration")
                base_ejection_duration = Duration(
                    unit=_base_ejection_duration.get("unit"),
                    value=_base_ejection_duration.get("value"),
                )
                if _interval is None:
                    raise MissingRequiredFieldError("interval")
                interval = Duration(
                    unit=_interval.get("unit"), value=_interval.get("value")
                )
                outlier_detection = OutlierDetection(
                    base_ejection_duration=base_ejection_duration,
                    interval=interval,
                    max_ejection_percent=_outlier_detection.get("maxEjectionPercent"),
                    max_server_errors=_outlier_detection.get("maxServerErrors"),
                )

            if _port_mapping is None:
                raise MissingRequiredFieldError("portMapping")
            port_mapping = PortMapping(
                port=_port_mapping.get("port"),
                protocol=_port_mapping.get("protocol"),
            )

            if _timeout is not None:
                _grpc_timeout = _timeout.get("grpc")
                _http_timeout = _timeout.get("http")
                _http2_timeout = _timeout.get("http2")
                _tcp_timeout = _timeout.get("tcp")
                grpc_timeout, http_timeout, http2_timeout, tcp_timeout = (
                    None,
                    None,
                    None,
                    None,
                )

                if _grpc_timeout is not None:
                    _idle = _grpc_timeout.get("idle")
                    _per_request = _grpc_timeout.get("perRequest")
                    idle, per_request = None, None
                    if _idle is not None:
                        idle = Duration(
                            unit=_idle.get("unit"), value=_idle.get("value")
                        )
                    if _per_request is not None:
                        per_request = Duration(
                            unit=_per_request.get("unit"),
                            value=_per_request.get("value"),
                        )
                    grpc_timeout = Timeout(idle=idle, per_request=per_request)
                if _http_timeout is not None:
                    _idle = _http_timeout.get("idle")
                    _per_request = _http_timeout.get("perRequest")
                    idle, per_request = None, None
                    if _idle is not None:
                        idle = Duration(
                            unit=_idle.get("unit"), value=_idle.get("value")
                        )
                    if _per_request is not None:
                        per_request = Duration(
                            unit=_per_request.get("unit"),
                            value=_per_request.get("value"),
                        )
                    http_timeout = Timeout(idle=idle, per_request=per_request)
                if _http2_timeout is not None:
                    _idle = _http2_timeout.get("idle")
                    _per_request = _http2_timeout.get("perRequest")
                    idle, per_request = None, None
                    if _idle is not None:
                        idle = Duration(
                            unit=_idle.get("unit"), value=_idle.get("value")
                        )
                    if _per_request is not None:
                        per_request = Duration(
                            unit=_per_request.get("unit"),
                            value=_per_request.get("value"),
                        )
                    http2_timeout = Timeout(idle=idle, per_request=per_request)
                if _tcp_timeout is not None:
                    _idle = _tcp_timeout.get("idle")
                    if _idle is None:
                        raise MissingRequiredFieldError("idle")
                    idle = Duration(unit=_idle.get("unit"), value=_idle.get("value"))
                    tcp_timeout = TCPTimeout(idle=idle)
                timeout = ProtocolTimeouts(
                    grpc=grpc_timeout,
                    http=http_timeout,
                    http2=http2_timeout,
                    tcp=tcp_timeout,
                )

            if _listener_tls is not None:
                _tls_listener_certificate = _listener_tls.get("certificate")
                _tls_listener_validation = _listener_tls.get("validation")
                tls_listener_validation = None
                if _tls_listener_certificate is None:
                    raise MissingRequiredFieldError("certificate")
                _listener_certificate_file = _tls_listener_certificate.get("file")
                _listener_certificate_sds = _tls_listener_certificate.get("sds")
                _listener_certificate_acm = _tls_listener_certificate.get("acm")
                (
                    listener_certificate_file,
                    listener_certificate_sds,
                    listener_certificate_acm,
                ) = None, None, None
                if _listener_certificate_file is not None:
                    listener_certificate_file = CertificateFileWithPrivateKey(
                        certificate_chain=_listener_certificate_file.get(
                            "certificateChain"
                        ),
                        private_key=_listener_certificate_file.get("privateKey"),
                    )
                if _listener_certificate_sds is not None:
                    listener_certificate_sds = SDS(
                        secret_name=_listener_certificate_sds.get("secretName")
                    )
                if _listener_certificate_acm is not None:
                    listener_certificate_acm = ListenerCertificateACM(
                        certificate_arn=_listener_certificate_acm.get("certificateArn")
                    )

                tls_listener_certificate = TLSListenerCertificate(
                    file=listener_certificate_file,
                    sds=listener_certificate_sds,
                    acm=listener_certificate_acm,
                )
                if _tls_listener_validation is not None:
                    _subject_alternative_names = _tls_listener_validation.get(
                        "subjectAlternativeNames"
                    )
                    _trust = _tls_listener_validation.get("trust")
                    subject_alternative_names = None
                    if _subject_alternative_names is not None:
                        _tls_listener_match = _subject_alternative_names.get("match")
                        tls_listener_match = VirtualNodeMatch(
                            exact=_tls_listener_match.get("exact")
                        )
                        subject_alternative_names = SubjectAlternativeNames(
                            match=tls_listener_match
                        )
                    if _trust is None:
                        raise MissingRequiredFieldError("trust")
                    _tls_listener_certificate_file = _trust.get("file")
                    _tls_listener_sds = _trust.get("sds")
                    tls_listener_certificate_file, tls_listener_sds = None, None
                    if _tls_listener_certificate_file is not None:
                        tls_listener_certificate_file = CertificateFile(
                            certificate_chain=_tls_listener_certificate_file.get(
                                "certificateChain"
                            )
                        )
                    if _tls_listener_sds is not None:
                        tls_listener_sds = SDS(
                            secret_name=_tls_listener_sds.get("secretName")
                        )
                    tls_listener_trust = Trust(
                        file=tls_listener_certificate_file, sds=tls_listener_sds
                    )

                    tls_listener_validation = TLSListenerValidation(
                        subject_alternative_names=subject_alternative_names,
                        trust=tls_listener_trust,
                    )
                listener_tls = ListenerTLS(
                    certificate=tls_listener_certificate,
                    mode=_listener_tls.get("mode"),
                    validation=tls_listener_validation,
                )

            listener = Listener(
                connection_pool=connection_pool,
                health_check=health_check,
                outlier_detection=outlier_detection,
                port_mapping=port_mapping,
                timeout=timeout,
                tls=listener_tls,
            )
            listeners.append(listener)

    if _logging is not None:
        _access_log = _logging.get("accessLog")
        access_log = None
        if _access_log is not None:
            _file = _access_log.get("file")
            file = None
            if _file is not None:
                _format = _file.get("format")
                format = None
                if _format is not None:
                    _json = _format.get("json")
                    json = None
                    if _json is not None:
                        json = list()
                        for item in _json:
                            json.append(
                                KeyValue(key=item.get("key"), value=item.get("value"))
                            )
                    format = LoggingFormat(json=json, text=_format.get("text"))
                file = AccessLogFile(format=format, path=_file.get("path"))
            access_log = AccessLog(file=file)
        logging = Logging(access_log=access_log)

    if _service_discovery is not None:
        _aws_cloud_map = _service_discovery.get("awsCloudMap")
        _dns = _service_discovery.get("dns")
        aws_cloud_map, dns = None, None
        if _aws_cloud_map is not None:
            _attributes = _aws_cloud_map.get("attributes")
            if _attributes is None:
                raise MissingRequiredFieldError("attributes")
            attributes = [
                KeyValue(key=attribute.get("key"), value=attribute.get("value"))
                for attribute in _attributes
            ]
            aws_cloud_map = AWSCloudMap(
                attributes=attributes,
                ip_preference=_aws_cloud_map.get("ipPreference"),
                namespace_name=_aws_cloud_map.get("namespaceName"),
                service_name=_aws_cloud_map.get("serviceName"),
            )
        if _dns is not None:
            dns = DNS(
                hostname=_dns.get("hostname"),
                ip_preference=_dns.get("ipPreference"),
                response_type=_dns.get("responseType"),
            )
        service_discovery = ServiceDiscovery(aws_cloud_map=aws_cloud_map, dns=dns)

    return VirtualNodeSpec(
        backend_defaults=backend_defaults,
        backends=backends,
        listeners=listeners,
        logging=logging,
        service_discovery=service_discovery,
    )



import math
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import camelcase_to_pascal, underscores_to_camelcase
from moto.utilities.paginator import paginate
from moto.utilities.utils import get_partition

from .exceptions import (
    AlreadyExistsException,
    ConflictException,
    ResourceNotFoundException,
    RulesetAlreadyExistsException,
    RulesetNotFoundException,
    ValidationException,
)


class DataBrewBackend(BaseBackend):
    PAGINATION_MODEL = {
        "list_recipes": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "name",
        },
        "list_recipe_versions": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "name",
        },
        "list_rulesets": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "name",
        },
        "list_datasets": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "name",
        },
        "list_jobs": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "name",
        },
    }

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.recipes: Dict[str, FakeRecipe] = OrderedDict()
        self.rulesets: Dict[str, FakeRuleset] = OrderedDict()
        self.datasets: Dict[str, FakeDataset] = OrderedDict()
        self.jobs: Dict[str, FakeJob] = OrderedDict()

    @staticmethod
    def validate_length(param: str, param_name: str, max_length: int) -> None:
        if len(param) > max_length:
            raise ValidationException(
                f"1 validation error detected: Value '{param}' at '{param_name}' failed to "
                f"satisfy constraint: Member must have length less than or equal to {max_length}"
            )

    def create_recipe(
        self,
        recipe_name: str,
        recipe_description: str,
        recipe_steps: List[Dict[str, Any]],
        tags: Dict[str, str],
    ) -> "FakeRecipeVersion":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateRecipe.html
        if recipe_name in self.recipes:
            raise ConflictException(f"The recipe {recipe_name} already exists")

        recipe = FakeRecipe(
            self.region_name, recipe_name, recipe_description, recipe_steps, tags
        )
        self.recipes[recipe_name] = recipe
        return recipe.latest_working

    def delete_recipe_version(self, recipe_name: str, recipe_version: str) -> None:
        if not FakeRecipe.version_is_valid(recipe_version, latest_published=False):
            raise ValidationException(
                f"Recipe {recipe_name} version {recipe_version} is invalid."
            )

        try:
            recipe = self.recipes[recipe_name]
        except KeyError:
            raise ResourceNotFoundException(f"The recipe {recipe_name} wasn't found")

        if (
            recipe_version != FakeRecipe.LATEST_WORKING
            and float(recipe_version) not in recipe.versions
        ):
            raise ResourceNotFoundException(
                f"The recipe {recipe_name} version {recipe_version} wasn't found."
            )

        if recipe_version in (
            FakeRecipe.LATEST_WORKING,
            str(recipe.latest_working.version),
        ):
            if recipe.latest_published is not None:
                # Can only delete latest working version when there are no others
                raise ValidationException(
                    f"Recipe version {recipe_version} is not allowed to be deleted"
                )
            else:
                del self.recipes[recipe_name]
        else:
            recipe.delete_published_version(recipe_version)

    def update_recipe(
        self,
        recipe_name: str,
        recipe_description: str,
        recipe_steps: List[Dict[str, Any]],
    ) -> None:
        if recipe_name not in self.recipes:
            raise ResourceNotFoundException(f"The recipe {recipe_name} wasn't found")

        recipe = self.recipes[recipe_name]
        recipe.update(recipe_description, recipe_steps)

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_recipes(
        self, recipe_version: Optional[str] = None
    ) -> List["FakeRecipeVersion"]:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListRecipes.html
        if recipe_version == FakeRecipe.LATEST_WORKING:
            version = "latest_working"
        elif recipe_version in (None, FakeRecipe.LATEST_PUBLISHED):
            version = "latest_published"
        else:
            raise ValidationException(
                f"Invalid version {recipe_version}. "
                "Valid versions are LATEST_PUBLISHED and LATEST_WORKING."
            )
        recipes = [getattr(self.recipes[key], version) for key in self.recipes]
        return [r for r in recipes if r is not None]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_recipe_versions(self, recipe_name: str) -> List["FakeRecipeVersion"]:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListRecipeVersions.html
        self.validate_length(recipe_name, "name", 255)

        recipe = self.recipes.get(recipe_name)
        if recipe is None:
            return []

        latest_working = recipe.latest_working

        recipe_versions = [
            recipe_version
            for recipe_version in recipe.versions.values()
            if recipe_version is not latest_working
        ]
        return [r for r in recipe_versions if r is not None]

    def describe_recipe(
        self, recipe_name: str, recipe_version: Optional[str] = None
    ) -> "FakeRecipeVersion":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_DescribeRecipe.html
        self.validate_length(recipe_name, "name", 255)

        if recipe_version is None:
            recipe_version = FakeRecipe.LATEST_PUBLISHED
        else:
            self.validate_length(recipe_version, "recipeVersion", 16)
            if not FakeRecipe.version_is_valid(recipe_version):
                raise ValidationException(
                    f"Recipe {recipe_name} version {recipe_version} isn't valid."
                )

        recipe = None
        if recipe_name in self.recipes:
            if recipe_version == FakeRecipe.LATEST_PUBLISHED:
                recipe = self.recipes[recipe_name].latest_published
            elif recipe_version == FakeRecipe.LATEST_WORKING:
                recipe = self.recipes[recipe_name].latest_working
            else:
                recipe = self.recipes[recipe_name].versions.get(float(recipe_version))
        if recipe is None:
            raise ResourceNotFoundException(
                f"The recipe {recipe_name} for version {recipe_version} wasn't found."
            )
        return recipe

    def publish_recipe(
        self, recipe_name: str, description: Optional[str] = None
    ) -> None:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_PublishRecipe.html
        self.validate_length(recipe_name, "name", 255)
        try:
            self.recipes[recipe_name].publish(description)
        except KeyError:
            raise ResourceNotFoundException(f"Recipe {recipe_name} wasn't found")

    def create_ruleset(
        self,
        ruleset_name: str,
        ruleset_description: str,
        ruleset_rules: List[Dict[str, Any]],
        ruleset_target_arn: str,
        tags: Dict[str, str],
    ) -> "FakeRuleset":
        if ruleset_name in self.rulesets:
            raise RulesetAlreadyExistsException()

        ruleset = FakeRuleset(
            self.region_name,
            ruleset_name,
            ruleset_description,
            ruleset_rules,
            ruleset_target_arn,
            tags,
        )
        self.rulesets[ruleset_name] = ruleset
        return ruleset

    def update_ruleset(
        self,
        ruleset_name: str,
        ruleset_description: str,
        ruleset_rules: List[Dict[str, Any]],
        tags: Dict[str, str],
    ) -> "FakeRuleset":
        if ruleset_name not in self.rulesets:
            raise RulesetNotFoundException(ruleset_name)

        ruleset = self.rulesets[ruleset_name]
        if ruleset_description is not None:
            ruleset.description = ruleset_description
        if ruleset_rules is not None:
            ruleset.rules = ruleset_rules
        if tags is not None:
            ruleset.tags = tags

        return ruleset

    def describe_ruleset(self, ruleset_name: str) -> "FakeRuleset":
        if ruleset_name not in self.rulesets:
            raise RulesetNotFoundException(ruleset_name)
        return self.rulesets[ruleset_name]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_rulesets(self) -> List["FakeRuleset"]:
        return list(self.rulesets.values())

    def delete_ruleset(self, ruleset_name: str) -> None:
        if ruleset_name not in self.rulesets:
            raise RulesetNotFoundException(ruleset_name)

        del self.rulesets[ruleset_name]

    def create_dataset(
        self,
        dataset_name: str,
        dataset_format: str,
        dataset_format_options: Dict[str, Any],
        dataset_input: Dict[str, Any],
        dataset_path_options: Dict[str, Any],
        tags: Dict[str, str],
    ) -> "FakeDataset":
        if dataset_name in self.datasets:
            raise AlreadyExistsException(dataset_name)

        dataset = FakeDataset(
            self.region_name,
            self.account_id,
            dataset_name,
            dataset_format,
            dataset_format_options,
            dataset_input,
            dataset_path_options,
            tags,
        )
        self.datasets[dataset_name] = dataset
        return dataset

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_datasets(self) -> List["FakeDataset"]:
        return list(self.datasets.values())

    def update_dataset(
        self,
        dataset_name: str,
        dataset_format: str,
        dataset_format_options: Dict[str, Any],
        dataset_input: Dict[str, Any],
        dataset_path_options: Dict[str, Any],
        tags: Dict[str, str],
    ) -> "FakeDataset":
        if dataset_name not in self.datasets:
            raise ResourceNotFoundException("One or more resources can't be found.")

        dataset = self.datasets[dataset_name]

        if dataset_format is not None:
            dataset.format = dataset_format
        if dataset_format_options is not None:
            dataset.format_options = dataset_format_options
        if dataset_input is not None:
            dataset.input = dataset_input
        if dataset_path_options is not None:
            dataset.path_options = dataset_path_options
        if tags is not None:
            dataset.tags = tags

        return dataset

    def delete_dataset(self, dataset_name: str) -> None:
        if dataset_name not in self.datasets:
            raise ResourceNotFoundException("One or more resources can't be found.")

        del self.datasets[dataset_name]

    def describe_dataset(self, dataset_name: str) -> "FakeDataset":
        if dataset_name not in self.datasets:
            raise ResourceNotFoundException("One or more resources can't be found.")

        return self.datasets[dataset_name]

    def describe_job(self, job_name: str) -> "FakeJob":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_DescribeJob.html
        self.validate_length(job_name, "name", 240)

        if job_name not in self.jobs:
            raise ResourceNotFoundException(f"Job {job_name} wasn't found.")

        return self.jobs[job_name]

    def delete_job(self, job_name: str) -> None:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_DeleteJob.html
        self.validate_length(job_name, "name", 240)

        if job_name not in self.jobs:
            raise ResourceNotFoundException(f"The job {job_name} wasn't found.")

        del self.jobs[job_name]

    def create_profile_job(self, **kwargs: Any) -> "FakeProfileJob":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateProfileJob.html
        job_name = kwargs["name"]
        self.validate_length(job_name, "name", 240)

        if job_name in self.jobs:
            raise ConflictException(
                f"The job {job_name} {self.jobs[job_name].job_type.lower()} job already exists."
            )

        job = FakeProfileJob(
            account_id=self.account_id, region_name=self.region_name, **kwargs
        )

        self.jobs[job_name] = job
        return job

    def create_recipe_job(self, **kwargs: Any) -> "FakeRecipeJob":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_CreateRecipeJob.html
        job_name = kwargs["name"]
        self.validate_length(job_name, "name", 240)

        if job_name in self.jobs:
            raise ConflictException(
                f"The job {job_name} {self.jobs[job_name].job_type.lower()} job already exists."
            )

        job = FakeRecipeJob(
            account_id=self.account_id, region_name=self.region_name, **kwargs
        )

        self.jobs[job_name] = job
        return job

    def update_job(self, **kwargs: Any) -> "FakeJob":
        job_name = kwargs["name"]
        self.validate_length(job_name, "name", 240)

        if job_name not in self.jobs:
            raise ResourceNotFoundException(f"The job {job_name} wasn't found")

        job = self.jobs[job_name]

        for param, value in kwargs.items():
            setattr(job, param, value)
        return job

    def update_recipe_job(self, **kwargs: Any) -> "FakeJob":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_UpdateRecipeJob.html
        return self.update_job(**kwargs)

    def update_profile_job(self, **kwargs: Any) -> "FakeJob":
        # https://docs.aws.amazon.com/databrew/latest/dg/API_UpdateProfileJob.html
        return self.update_job(**kwargs)

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_jobs(
        self, dataset_name: Optional[str] = None, project_name: Optional[str] = None
    ) -> List["FakeJob"]:
        # https://docs.aws.amazon.com/databrew/latest/dg/API_ListJobs.html
        if dataset_name is not None:
            self.validate_length(dataset_name, "datasetName", 255)
        if project_name is not None:
            self.validate_length(project_name, "projectName", 255)

        def filter_jobs(job: FakeJob) -> bool:
            if dataset_name is not None and job.dataset_name != dataset_name:
                return False
            if (
                project_name is not None
                and getattr(job, "project_name", None) != project_name
            ):
                return False
            return True

        return list(filter(filter_jobs, self.jobs.values()))


class FakeRecipe(BaseModel):
    INITIAL_VERSION = 0.1
    LATEST_WORKING = "LATEST_WORKING"
    LATEST_PUBLISHED = "LATEST_PUBLISHED"

    @classmethod
    def version_is_valid(
        cls, version: str, latest_working: bool = True, latest_published: bool = True
    ) -> bool:
        validity = True

        if len(version) < 1 or len(version) > 16:
            validity = False
        else:
            try:
                float(version)
            except ValueError:
                if not (
                    (version == cls.LATEST_WORKING and latest_working)
                    or (version == cls.LATEST_PUBLISHED and latest_published)
                ):
                    validity = False
        return validity

    def __init__(
        self,
        region_name: str,
        recipe_name: str,
        recipe_description: str,
        recipe_steps: List[Dict[str, Any]],
        tags: Dict[str, str],
    ):
        self.versions: Dict[float, FakeRecipeVersion] = OrderedDict()
        self.versions[self.INITIAL_VERSION] = FakeRecipeVersion(
            region_name,
            recipe_name,
            recipe_description,
            recipe_steps,
            tags,
            version=self.INITIAL_VERSION,
        )
        self.latest_working = self.versions[self.INITIAL_VERSION]
        self.latest_published: Optional[FakeRecipeVersion] = None

    def publish(self, description: Optional[str] = None) -> None:
        self.latest_published = self.latest_working
        self.latest_working = deepcopy(self.latest_working)
        self.latest_published.publish(description)
        del self.versions[self.latest_working.version]
        self.versions[self.latest_published.version] = self.latest_published
        self.latest_working.version = self.latest_published.version + 0.1

        if self.latest_published.published_date:
            self.latest_working.created_time = self.latest_published.published_date
        self.versions[self.latest_working.version] = self.latest_working

    def update(
        self, description: Optional[str], steps: Optional[List[Dict[str, Any]]]
    ) -> None:
        if description is not None:
            self.latest_working.description = description
        if steps is not None:
            self.latest_working.steps = steps

    def delete_published_version(self, version: str) -> None:
        float_version = float(version)
        assert float_version.is_integer()
        if float_version == self.latest_published.version:  # type: ignore[union-attr]
            prev_version = float_version - 1.0
            # Iterate back through the published versions until we find one that exists
            while prev_version >= 1.0:
                if prev_version in self.versions:
                    self.latest_published = self.versions[prev_version]
                    break
                prev_version -= 1.0
            else:
                # Didn't find an earlier published version
                self.latest_published = None
        del self.versions[float_version]


class FakeRecipeVersion(BaseModel):
    def __init__(
        self,
        region_name: str,
        recipe_name: str,
        recipe_description: str,
        recipe_steps: List[Dict[str, Any]],
        tags: Dict[str, str],
        version: float,
    ):
        self.region_name = region_name
        self.name = recipe_name
        self.description = recipe_description
        self.steps = recipe_steps
        self.created_time = datetime.now()
        self.tags = tags
        self.published_date: Optional[datetime] = None
        self.version = version

    def as_dict(self) -> Dict[str, Any]:
        dict_recipe = {
            "Name": self.name,
            "Steps": self.steps,
            "Description": self.description,
            "CreateDate": f"{self.created_time.timestamp():.3f}",
            "Tags": self.tags or dict(),
            "RecipeVersion": str(self.version),
        }
        if self.published_date is not None:
            dict_recipe["PublishedDate"] = f"{self.published_date.timestamp():.3f}"

        return dict_recipe

    def publish(self, description: Optional[str]) -> None:
        self.version = float(math.ceil(self.version))
        self.published_date = datetime.now()
        if description is not None:
            self.description = description


class FakeRuleset(BaseModel):
    def __init__(
        self,
        region_name: str,
        ruleset_name: str,
        ruleset_description: str,
        ruleset_rules: List[Dict[str, Any]],
        ruleset_target_arn: str,
        tags: Dict[str, str],
    ):
        self.region_name = region_name
        self.name = ruleset_name
        self.description = ruleset_description
        self.rules = ruleset_rules
        self.target_arn = ruleset_target_arn
        self.created_time = datetime.now()

        self.tags = tags

    def as_dict(self) -> Dict[str, Any]:
        return {
            "Name": self.name,
            "Rules": self.rules,
            "Description": self.description,
            "TargetArn": self.target_arn,
            "CreateDate": f"{self.created_time.timestamp():.3f}",
            "Tags": self.tags or dict(),
        }


class FakeDataset(BaseModel):
    def __init__(
        self,
        region_name: str,
        account_id: str,
        dataset_name: str,
        dataset_format: str,
        dataset_format_options: Dict[str, Any],
        dataset_input: Dict[str, Any],
        dataset_path_options: Dict[str, Any],
        tags: Dict[str, str],
    ):
        self.region_name = region_name
        self.account_id = account_id
        self.name = dataset_name
        self.format = dataset_format
        self.format_options = dataset_format_options
        self.input = dataset_input
        self.path_options = dataset_path_options
        self.created_time = datetime.now()
        self.tags = tags

    @property
    def resource_arn(self) -> str:
        return f"arn:{get_partition(self.region_name)}:databrew:{self.region_name}:{self.account_id}:dataset/{self.name}"

    def as_dict(self) -> Dict[str, Any]:
        return {
            "Name": self.name,
            "Format": self.format,
            "FormatOptions": self.format_options,
            "Input": self.input,
            "PathOptions": self.path_options,
            "CreateDate": f"{self.created_time.timestamp():.3f}",
            "Tags": self.tags or dict(),
            "ResourceArn": self.resource_arn,
        }


class BaseModelABCMeta(ABCMeta, type(BaseModel)):  # type: ignore[misc]
    pass


class FakeJob(BaseModel, metaclass=BaseModelABCMeta):  # type: ignore[misc]
    ENCRYPTION_MODES = ("SSE-S3", "SSE-KMS")
    LOG_SUBSCRIPTION_VALUES = ("ENABLE", "DISABLE")

    @property
    @abstractmethod
    def local_attrs(self) -> List[str]:
        raise NotImplementedError

    def __init__(self, account_id: str, region_name: str, **kwargs: Any):
        self.account_id = account_id
        self.region_name = region_name
        self.name = kwargs.get("name")
        self.created_time = datetime.now()
        self.dataset_name = kwargs.get("dataset_name")
        self.encryption_mode = kwargs.get("encryption_mode")
        self.log_subscription = kwargs.get("log_subscription")
        self.max_capacity = kwargs.get("max_capacity")
        self.max_retries = kwargs.get("max_retries")
        self.role_arn = kwargs.get("role_arn")
        self.tags = kwargs.get("tags")
        self.validate()
        # Set attributes specific to subclass
        for k in self.local_attrs:
            setattr(self, k, kwargs.get(k))

    def validate(self) -> None:
        if self.encryption_mode is not None:
            if self.encryption_mode not in FakeJob.ENCRYPTION_MODES:
                raise ValidationException(
                    f"1 validation error detected: Value '{self.encryption_mode}' at 'encryptionMode' failed to satisfy constraint: Member must satisfy enum value set: [{', '.join(self.ENCRYPTION_MODES)}]"
                )
        if self.log_subscription is not None:
            if self.log_subscription not in FakeJob.LOG_SUBSCRIPTION_VALUES:
                raise ValidationException(
                    f"1 validation error detected: Value '{self.log_subscription}' at 'logSubscription' failed to satisfy constraint: Member must satisfy enum value set: [{', '.join(self.LOG_SUBSCRIPTION_VALUES)}]"
                )

    @property
    @abstractmethod
    def job_type(self) -> str:
        pass

    @property
    def resource_arn(self) -> str:
        return f"arn:{get_partition(self.region_name)}:databrew:{self.region_name}:{self.account_id}:job/{self.name}"

    def as_dict(self) -> Dict[str, Any]:
        rtn_dict = {
            "Name": self.name,
            "AccountId": self.account_id,
            "CreateDate": f"{self.created_time.timestamp():.3f}",
            "DatasetName": self.dataset_name,
            "EncryptionMode": self.encryption_mode,
            "Tags": self.tags or dict(),
            "LogSubscription": self.log_subscription,
            "MaxCapacity": self.max_capacity,
            "MaxRetries": self.max_retries,
            "ResourceArn": self.resource_arn,
            "RoleArn": self.role_arn,
            "Type": self.job_type,
        }

        # Add in subclass attributes
        for k in self.local_attrs:
            rtn_dict[camelcase_to_pascal(underscores_to_camelcase(k))] = getattr(
                self, k
            )

        # Remove items that have a value of None
        rtn_dict = {k: v for k, v in rtn_dict.items() if v is not None}

        return rtn_dict


class FakeProfileJob(FakeJob):  # type: ignore[misc]
    job_type = "PROFILE"
    local_attrs = ["output_location", "configuration", "validation_configurations"]


class FakeRecipeJob(FakeJob):  # type: ignore[misc]
    local_attrs = [
        "database_outputs",
        "data_catalog_outputs",
        "outputs",
        "project_name",
        "recipe_reference",
    ]
    job_type = "RECIPE"


databrew_backends = BackendDict(DataBrewBackend, "databrew")
